!+
! Module: emevents_mod
!
! Authors:  B. Thooris, J. Ballet, M. Lortholary - SAp  CEA Saclay
!          
! Description: Analyse event information from one CCD/node
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module emevents_mod

  use emutils_mod
  use edusoft_mod
  use sla
  use eventAttributes
  use cal
  use dal
  use param
  use errorHandling
  
  implicit none
  
  private 

! Names of the task parameters (2nd string for the keywords)

  character(len=*), parameter :: inevent     = "odfeventset", &
                                 newoutput   = "newoutput", &
                                 outevent    = "eventset", &
                                 withframe   = "withframeset", &
                                 framefile   = "frameset", &
                                 withoffvar  = "withoffvarsets", &
                                 offvarfile  = "offvarsets", &
                                 twonodes    = "withothernode", &
                                 inevent2    = "othereventset", &
                                 framefile2  = "otherframeset", &
                                 evrec_on    = "analysepatterns", &
                                 cutbad_on   = "flagbadpixels", &
                                 diago_on    = "splitdiagonals", &
                                 keepsifluor = "keepsifluor", &
                                 badrows_on  = "rejectrows", &
                                 threshrow   = "maxeventsperrow", &
                                 bade3_on    = "rejectbade3", &
                                 e3threshold = "e3thresholds", &
                                 puttime_on  = "settimes", &
                                 cutgti_on   = "flagbadtimes", &
                                 spgatti_on  = "flagtruncatede1", &
                                 writegatti  = "writegattispecset", &
                                 gattispec   = "gattispecset", &
                                 blocksize   = "blocksize", &
                                 randomxy    = "randomizeposition", &
                                 randomtime  = "randomizetime", &
                                 frames_on   = "checkframes"

! Name of the extensions

  character(len=*), parameter :: frame_ext="FRAMES", event_ext="EVENTS", &
                                 expos_ext="EXPOSURE", badpix_ext="BADPIX",&
                                 offset_ext="OFFSETS"

! Name of the columns

  character(len=8), parameter :: fram = "FRAME   ", time = "TIME    ", &
                                 nval = "NVALID  ", crra = "CRRATIO ", &
                                 flag = "FLAG    ", gatv = "GATTIVAL", &
                                 rawx = "RAWX    ", rawy = "RAWY    ", &
                                 ene1 = "ENERGYE1", ene2 = "ENERGYE2", &
                                 ene3 = "ENERGYE3", ene4 = "ENERGYE4", &
                                 patt = "PATTERN ", peri = "PERIPIX ", &
                                 detx = "DETX    ", dety = "DETY    ", &
                                 skyx = "X       ", skyy = "Y       ", &
                                  pha = "PHA     ",   pi = "PI      ", &
                                 en12 = "ENGYE1E2"

! Units for the new columns

  character(len=16), parameter :: coord_unit = "0.05 ARCSECONDS ",  &
                                  ener_unit  = "CHAN            ",  &
                                  time_unit  = "s               ",  &
                                  angl_unit  = "deg             "

! Units in arcsecs and offset (to ensure positive coordinates) if any
! for detector coordinates

  real(double),        parameter :: det_unit   = 0.05
  integer(kind=int16), parameter :: det_offset = 0
  integer(kind=int32), parameter :: detSysType = DETECTOR_OPTICSCENTER_SYSTEM

! Threshold on frame flag (above --> bad frames, below --> spoiled frames)

  integer(kind=int16), parameter :: flag_reject =   64,  &
                                    flag_hkgti  =  256

! Limit on fraction of good events below which a warning is sent

  real, parameter :: fracmin = 0.1

! Maximum PHA channel

  integer, parameter :: maxchan = 4095

! Routine name

  character(len=*), parameter :: taskname = "emevents"

! Global variables
! Column handles

  type(ColumnT)  :: fr_col, time_col, gatv_col, flag_col, cr_col, &
                    frev_col, rawx_col, rawy_col, patt_col, e1_col, flev_col

! Columns of data
! nvalid is int32 to accommodate extreme cases

  integer(kind=int8),  dimension(:), pointer   :: pattern
  integer(kind=int16), dimension(:), pointer   :: quality, gattival, &
                                                  posx, posy, ener1, quality2
  integer(kind=int32), dimension(:), pointer   :: frameno, frameev, qualev,  &
                                                  nvalid, nvalid2
  real(kind=single),   dimension(:), pointer   :: crratio
  real(kind=double),   dimension(:), pointer   :: timetag
  integer(kind=int16), dimension(:,:), pointer :: offsetX, offsetY

! Function flags
! onevrec   : Should EV_REC   be applied ?
! oncutbad  : Should CUT_BAD  be applied ?
! ondiago   : Should DIAGO    be applied ?
! onsifluor : Should fluorescence diagonals stick together ?
! oncutgti  : Should CUT_GTI  be applied ?
! onputxy   : Should PUT_XY   be applied ?
! rowreject : Should REJECT_ROWS be applied ?
! e3reject  : Should REJECT_E3 be applied ?
! onputtime : Should PUT_TIME be applied ?
! onspgatti : Should SP_GATTI be applied ?
! xyrandom  : Should DETX/DETY be randomized ?
! timerandom : Should TIME be randomized ?
! onbinodal : Should data from the other node be merged ?
! overevent : Do we overwrite the input event file  ?
! fromodf   : Is the input file straight from the ODF ?
! onoffvar  : Do we know the offset/variance file ?
! outgatti  : Do we write the GATTI histograms to a file ?
! onexpos   : Is there a CRRATIO column in input ?
! frameread : Is an input frame file available ?
! frameflag : Is there a FLAG column in the frame file ?
! frameflag2: Idem for 2nd node (binodal)
! nonIncreasingFrame : Is there an instance of non-increasing frame number ?
! nonIncreasingFrame2 : Idem for 2nd node (binodal)
! imageMode : Is this IMAGING mode data ?
! timeMode  : Is this TIMING mode data ?
! rimagMode : Is this REDUCED IMAGING mode data ?
! ctimeMode : Is this COMPRESSED TIMING mode data ?

  logical :: oncutbad, ondiago, oncutgti, onputxy, onputtime, onspgatti,  &
             onevrec, onbinodal, onoffvar, onexpos, rowreject, e3reject,  &
             overevent, fromodf, frameread, frameflag, outgatti, onsifluor, &
             xyrandom, timerandom, nonIncreasingFrame, nonIncreasingFrame2, &
             imageMode, timeMode,rimagMode, ctimeMode, frameflag2

! frameoff   : Frame number minus index (starting at 0) in frame table
! frameoff2  : Idem for 2nd node (binodal)
! firstframe : Number of the first frame with events in it
! lastframe  : Number of the last frame with events in it
! offframe   : Offset between 1st frame input number and the second one
!              (binodal) defined on the basis of the time information
! nmaxevents : Size of event buffer for reanalyze
! maxevrow   : Maximum number of events per row/frame
! e3min      : Lower threshold on E3

  integer :: frameoff, firstframe, lastframe, nmaxevents, maxevrow, e3min
  integer :: frameoff2, offframe

! Maximum size of the read-out area
! Under/overscans (to be read in CAL)

  integer :: underx, overx, undery, overy
  integer :: x_size_window, y_size_window
  integer :: x_size_window_bin

  integer :: ccdnode

  public emevents

contains 

  !---------------------------------------------------------------------
  subroutine emevents
  !---------------------------------------------------------------------

    character(len=1024) :: event_out, frame_in, gattifile, event2, frame2
    integer :: neventsin

! Declarations for the DAL

    type(TableT)   :: ev1_tab, ev_tab, fr_tab, off_tab

! Read parameters

    call getParamValues(ev1_tab, neventsin, off_tab, event_out, frame_in,  &
                    event2, frame2, gattifile)

    call openEventsOut(event_out, ev1_tab, ev_tab)

    if (frameread) then
       call readFrames(frame_in, ev1_tab, ev_tab, fr_tab)
    else
       call putPrimaryKeywords(block(ev_tab))  ! Normally done via frame file
       call getNvalid(ev1_tab, frameno, nvalid, firstframe, lastframe,  &
                      frameoff, nonIncreasingFrame)
    endif

    call initCAL(ev_tab)

    call addOffsets(ev_tab, off_tab)           ! Needs the CAL

    if (fromodf)  then
       call reanalyze(ev1_tab, ev_tab, event2, frame2)
       call release(parent(ev1_tab))
    endif

    call readColumns(ev_tab)

    if (oncutgti)   call cutGTI(ev_tab, fr_tab)

    if (onputxy) then
       call putXY(ev_tab)
       call putDetCoords(ev_tab)
    endif

    if (onputtime)  call putTime(ev_tab)

    if (rowreject)  call rejectRows(ev_tab)

    if (e3reject) call rejectE3(ev_tab)

! Truncated events should be rejected in free run mode too
! Offsets only, GATTI=0

    if (onspgatti)  call spGatti(ev_tab, gattifile)

    call checkRejected(ev_tab, neventsin)

    if (onoffvar) then
       call release(off_tab)
    else if (oncutbad .or. onbinodal .or. rimagMode) then
       deallocate(offsetX,offsetY)
    endif

    call release_columns
    if (frameread)  call release(parent(fr_tab))
    call release(parent(ev_tab))

  end subroutine emevents

! Reanalyze all events by projecting them onto the CCD, cutting out
! bad pixels and running the EDUsoft.
! This routine loops over frames.
! This routine changes the number of events and creates them incrementally.
! It assumes the events are isolated in input (as created by the EDU),
! but does not preserve that in output. So cannot be run twice. 

  subroutine reanalyze(ev1_tab, ev_tab, evfile2, frfile2)

! ev1_tab   : Handle to the input event table
! ev_tab    : Handle to the output event table
! evfile2   : Name of the second input event file (BINODAL)
! frfile2   : Name of the second input frame file (BINODAL)


    character(len=*), intent(in)   :: evfile2, frfile2
    type(TableT), intent(in)       :: ev1_tab
    type(TableT), intent(inout)    :: ev_tab

! Variables for the DAL

    type(TableT)   :: bad1_tab, ev2_tab, fr2_tab
    type(ColumnT)  :: e2_col, e3_col, e4_col, peri_col

    integer(kind=int8),  dimension(:), pointer :: periout
    integer(kind=int16), dimension(:), pointer :: e2out, e3out, e4out
    real(kind=single),   dimension(:), pointer :: phaout
    integer(kind=int8),  dimension(:,:,:), pointer :: patterns, patterns_tim
    integer(kind=int16)                        :: eduThreshold, threshold
    integer(kind=int16), dimension(:), pointer :: badx, bady, badtype,  &
                                                  badext, badflag
    integer(kind=int32), dimension(:), allocatable :: bad_map
    integer(kind=int32)                        :: curflag

! Temporary storage of events

    type(sas_evt),  dimension(:), allocatable :: evtbuf, evtbuf2
    integer(kind=int32), dimension(:), allocatable :: orderbuf, orderbuf2

    integer :: windX1, windY1, emdhlow, emdhupp, edumode
    integer :: nevents, nevents2, nevents3, nevents_n2, nevents2_n2
    integer :: nvmax, ncray
    integer :: noutside, ndiago, nbelow, nabove, nbad, jbad, ninside2
    integer :: event0, event2, order, irow, jrow, ix, event0_n2
    integer :: frame0, frame1
    integer :: emini, emaxi, offpattern
    logical, dimension(:), pointer :: diagonal
    integer(kind=int32) :: timemodeval
    integer :: edumodeval

    call message(AppMsg,Verbose,"Entering REANALYZE ...")

! Declare the columns in the output file
! Write TLMIN/TLMAX keywords

    call declare_all_columns(ev_tab,e2_col,e3_col,e4_col,peri_col)
    call release_all_columns(e2_col,e3_col,e4_col,peri_col)
    call message(AppMsg,Noisy,"Output columns declared ...")

    if (numberOfRows(ev1_tab) < 1) then
       call add_new_columns(ev_tab)
       return
    endif

! In timing mode, full_frame mode is faked (for CAL and edumode)
! but timing patterns are got for cutBad
! Get and store the current mode from CAL
! Then change the mode to Full Frame
! For edumode the same operation is performed after initEdusoft()

    if (timeMode) then
       call CAL_getState( ccdModeId=timemodeval )
       call CAL_getEventPatterns(patterns_tim, eduThreshold)
       call CAL_setState( ccdModeId=PRIME_FULL_WINDOW )
       edumodeval = edumode
       edumode = 3  ! Imagery EDU mode       
    endif

! Read the pattern library

    if (.not. ctimeMode) then
       call CAL_getEventPatterns(patterns, eduThreshold)
       call initEdusoft(ev1_tab, threshold, edumode)
    endif

! Fake imagery mode

    if (timeMode) then
       edumodeval = edumode
       edumode = 3  ! Imagery EDU mode       
    endif

! Prepare DIAGO

    if (ondiago) then
       call diagonalPattern(patterns, diagonal)   ! Get diagonal events
       if (onsifluor)  &
          call initSifluor(patterns, diagonal, emini, emaxi, offpattern)
    endif

! Read the BADPIX extension
! Type: 0=OK, 1=hot, 2=flickering, 3=dead, 4=pinhole, 5=anomalous
! Flag: 1=uplinked, 2=in CCF but not uplinked, 3=new

    if (oncutbad) then
       bad1_tab = table(parent(ev1_tab),badpix_ext)
       call readBadpix(bad1_tab, badx, bady, badtype, badext, badflag)
       if (timeMode) then
          allocate(bad_map(x_size_window))
          bad_map = 0
          do jbad = 0,size(badx)-1
             select case(int(badtype(jbad),kind=int32))
             case (INTACT)
                curflag = 0
             case (DEAD)
                curflag = 0
             case default
                curflag = EVATT_ON_BADPIX
             end select

! The event coordinates are still offset by 2 + the underscan

             bad_map(underx+1+badx(jbad)) = curflag
          enddo
          write(errstr,*) "CUT_BAD: ", sum(bad_map/EVATT_ON_BADPIX),  &
                          " bad pixels will be removed"
       else
          write(errstr,*) "CUT_BAD: ", sum(int(badext)),  &
                          " bad pixels will be removed"
       endif
       call message(AppMsg,Verbose)
    endif

! Read the EMDH thresholds

    emdhlow = int32Attribute(ev1_tab,"EMDHLOW")
    emdhupp = int32Attribute(ev1_tab,"EMDHUPP")
    if (emdhlow > emdhupp) then
       write(errstr,*) "EMDHUPP =", emdhupp, " should be larger than ",  &
                       "EMDHLOW =", emdhlow
       call warning("reanalyze11")
    else if (onevrec .and. threshold > emdhupp) then
       write(errstr,*) "EMDHUPP =", emdhupp, " should be larger than ",  &
                       "EDUTHR =", threshold
       call warning("reanalyze12")
    else
       write(errstr,*) "EMDHLOW =", emdhlow, "     EMDHUPP =", emdhupp
       call message(AppMsg,Verbose)
    endif

! Increase the buffer size in case of very long frames

    nvmax = maxval(nvalid)

! In 2-node mode, open secondary frame and events files
! Establish the frame correspondance in binodal case between the 2 nodes
! on the basis of the time information (by computing offframe)
! onbinodal may become False if openEvents2In or read2Frames end in error

    if (onbinodal) call openEvents2In(evfile2, ev1_tab, ev2_tab)
    if (onbinodal) then
       call read2Frames(frfile2, ev2_tab, fr2_tab)
       if (onbinodal) then
          nvmax = max(nvmax,int(maxval(nvalid2)))
       else
          call release(parent(ev2_tab))
       endif
    endif

    if (nmaxevents < nvmax) then
       write(errstr,*) "The size of the events buffer ", nmaxevents,  &
            " is smaller than the maximum number of events in a frame",  &
            nvmax, ". ", blocksize, " is increased."
       call warning("reanalyze15")
       nmaxevents = nvmax
    endif

! Allocate the buffer arrays

    allocate(evtbuf(2*nmaxevents),orderbuf(2*nmaxevents))
    if (onbinodal) then
         allocate(evtbuf2(2*nmaxevents),orderbuf2(2*nmaxevents))
    else
         allocate(evtbuf2(2),orderbuf2(2)) ! Fake Allocation for cutBad() call
    endif

! Loop on blocks of frames

    frame0 = firstframe
    event0 = 0
    event2 = 0
    noutside = 0
    ninside2 = 0
    nbad = 0
    ndiago = 0
    nbelow = 0
    nabove = 0
    windX1 = x_size_window - overx + 2 - 1
    windY1 = y_size_window - overy + 2 - 1

    if (onbinodal) then

! event0_n2 is the subscript in evtbuf2 of the 1st event the frame of which
! is frame0+offframe, it can be calculated from nvalid2

       event0_n2 = 0
       if ( frame0+offframe-frameoff2-1 >= 0 ) &
            event0_n2 = sum(int(nvalid2(0:frame0+offframe-frameoff2-1)))

    endif

    do while (frame0 <= lastframe)

! Find last frame such that nevents <= nmaxevents

       if ( .not. onbinodal ) then      
          nevents = 0
          do frame1 = frame0,lastframe
             nevents = nevents + nvalid(frame1-frameoff)
             if (nevents > nmaxevents) then
                nevents = nevents - nvalid(frame1-frameoff)
                exit
             endif
          enddo
          frame1 = frame1 - 1

       else 

! In binodal the calculation of frame1 takes the 2 lists into account

          nevents = 0
          nevents_n2 = 0

          do frame1 = frame0,lastframe
             nevents = nevents + nvalid(frame1-frameoff)
             nevents_n2 = nevents_n2 + nvalid2(frame1+offframe-frameoff2)
             if ((nevents > nmaxevents) .or. (nevents_n2 > nmaxevents)) then
                nevents = nevents - nvalid(frame1-frameoff)
                nevents_n2 = nevents_n2 - nvalid2(frame1+offframe-frameoff2)
                exit
             endif
          enddo
          frame1 = frame1 - 1

       endif

       call getEvents(nevents, event0, ev1_tab, evtbuf)
       orderbuf(1:nevents) = (/(order,order=1,nevents)/)
       nevents2 = nevents

! Weed out events with non-increasing frame number

       if (nonIncreasingFrame)  &
            call checkEventOrder(frame0, frame1, frameoff, evtbuf, nvalid,  &
                                 orderbuf, nevents2)

! Binodal

       if (onbinodal) then
          nevents_n2 = &
         sum(int(nvalid2(frame0+offframe-frameoff2:frame1+offframe-frameoff2)))

          call getEvents(nevents_n2, event0_n2, ev2_tab, evtbuf2)
          orderbuf2(1:nevents_n2) = (/(order,order=1,nevents_n2)/)
          nevents2_n2 = nevents_n2

          if (nonIncreasingFrame2)  &
               call checkEventOrder(frame0+offframe, frame1+offframe,  &
                          frameoff2, evtbuf2, nvalid2, orderbuf2, nevents2_n2)
       endif

! Check that no event has invalid coordinates or PATTERN
! Note that in timing mode events with rawy=0 or 1 are possible 

       if (imageMode .or. rimagMode) then
          if ( .not. onbinodal ) then
             nevents3 = 0
             do jrow = 1,nevents2
                call checkValid(jrow,evtbuf,frameoff,2,x_size_window-1, &
                                0,y_size_window-1,0,size(patterns)-1,  &
                                .true.,orderbuf,nvalid,nevents3,noutside)
             enddo
             nevents2 = nevents3
          else

! On binodal, check evtbuf()%rawx < x_size_window/2 and idem for evtbuf2()
! Moreover, only events with rawx >= x_size_window_bin-overx ( 305 = 303 + 2 )
! are selected here for 2nd node.
! (the last physical pixel in binodal is Xccd=304 ie rawx=306) events with
! rawx=306 or 305 which could contribute to pixel on the central edge are
! accepted.

             nevents3 = 0
             do jrow = 1,nevents2
                call checkValid(jrow,evtbuf,frameoff,2,x_size_window_bin-1, &
                                2,y_size_window-1,0,size(patterns)-1,  &
                                .true.,orderbuf,nvalid,nevents3,noutside)
             enddo
             nevents2 = nevents3

! 2nd node 

             nevents3 = 0
             do jrow = 1,nevents2_n2
                call checkValid(jrow,evtbuf2,frameoff2,  &
                                x_size_window_bin-overx,x_size_window_bin-1, &
                                2,y_size_window-1,0,size(patterns)-1,  &
                                .true.,orderbuf2,nvalid2,nevents3)
             enddo
             ninside2 = ninside2 + nevents3
          endif

       else if (timeMode .and. onevrec) then
          nevents3 = 0
          do jrow = 1,nevents2
             call checkValid(jrow,evtbuf,frameoff,2,x_size_window-1, &
                             0,y_size_window-1,0,size(patterns)-1,  &
                             .true.,orderbuf,nvalid,nevents3,noutside)
          enddo
          nevents2 = nevents3
       endif

       if (oncutbad .or. onevrec) then
          if (imageMode .or. rimagMode .or. onevrec ) then

             if (.not. timeMode) then
                call cutBad(frame0,frame1,edumode,patterns,badx,bady,badext,  &
                            threshold,orderbuf,evtbuf,orderbuf2,evtbuf2,irow)
             else
                call cutBad(frame0,frame1,edumode,patterns,badx,bady,badext,  &
                            threshold,orderbuf,evtbuf,orderbuf2,evtbuf2,irow, &
                            patterns_tim)
             endif


             nbad = nbad + irow
             nevents2 = sum(int(nvalid(frame0-frameoff:frame1-frameoff)))

! If TIMING mode simply flag events on bad pixels

          else if (timeMode .and. .not.onevrec) then
             do jrow = 1,nevents2
                order = orderbuf(jrow)
                ix = evtbuf(order)%rawx
                if (ix >= -1 .and. ix <= x_size_window)  then
                   if (bad_map(ix) > 0) then
                      evtbuf(order)%flag = ior(evtbuf(order)%flag,bad_map(ix))
                      nbad = nbad + 1
                   endif
                endif
             enddo
          endif
       endif

! orderbuf indicates which events are kept in which order

       if (ondiago) then
          call diago(nevents2, patterns, diagonal, emini, emaxi, offpattern, &
                     orderbuf, evtbuf, irow)
          ndiago = ndiago + irow
       endif

! remove events which are located in the opposite half of the CCD

       if ( onbinodal ) then
          call cleanhalf( nevents2, orderbuf, evtbuf )
       endif

                                
! Flag all events below EMDH lower threshold
! Flag all events above EMDH upper threshold

       do jrow = 1,nevents2
          order = orderbuf(jrow)          
          if (evtbuf(order)%e1+evtbuf(order)%e2 > emdhupp) then
             evtbuf(order)%flag = ior(evtbuf(order)%flag,  &
                                      EVATT_OUTSIDE_THRESHOLDS)
             nabove = nabove + 1
          else if (evtbuf(order)%e1+evtbuf(order)%e2 < emdhlow) then
             evtbuf(order)%flag = ior(evtbuf(order)%flag,  &
                                      EVATT_OUTSIDE_THRESHOLDS)
             nbelow = nbelow + 1
          endif
       enddo

! Write the buffer's contents to the output file
! No randomisation of E1 should take place before SP_GATTI

       call insertRows(ev_tab,event2,nevents2)
       call enter_all_columns(ev_tab,e2_col,e3_col,e4_col,peri_col,  &
                              periout,e2out,e3out,e4out,phaout)
       frameev(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%frame
          posx(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%rawx
          posy(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%rawy
        qualev(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%flag
       pattern(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%pattern
       periout(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%peripix
       if (imageMode .or. rimagMode) then
          ener1(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%e1
          e2out(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%e2
          if (imageMode) then
             e3out(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%e3
             e4out(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%e4
          endif
       else if (timeMode .or. ctimeMode) then
          phaout(event2:event2+nevents2-1) = evtbuf(orderbuf(1:nevents2))%e1
       endif
       call release_all_columns(e2_col,e3_col,e4_col,peri_col)

       frame0 = frame1 + 1
       event0 = event0 + nevents
       event2 = event2 + nevents2
       if (onbinodal) then
          event0_n2 = event0_n2 + nevents_n2
       endif

       write(errstr,*) "event number ",event0
       call message(AppMsg,Verbose)

    enddo  ! end of loop on blocks of frames


! In timing mode restore the mode in the CAL and edumode

    if (timeMode) then
       call CAL_setState( ccdModeId=timemodeval )  
       edumode = edumodeval

       ncray = sum(iand(qualev,EVATT_COSMIC_RAY)/EVATT_COSMIC_RAY)
       write(errstr,*) ncray, " events identified as cosmic-rays"
       call message(AppMsg,Sparse)
    endif

    if (timeMode .and. oncutbad) deallocate(bad_map)

    if (ondiago) then
       write(errstr,*) ndiago, " diagonal events split in two"
       call message(AppMsg,Sparse)
    endif

    if (onbinodal) then
       write(errstr,*) ninside2, &
                       " events from 2nd node were considered for merging"
       call message(AppMsg,Sparse)
    endif

    if (oncutbad .or. rimagMode) then

       if (noutside > 0) then
          write(errstr,*) noutside, " events with invalid coordinates",  &
                          " or PATTERN"
          call message(AppMsg,Sparse)
       endif

       write(errstr,*) nbad, " events removed or modified by CUT_BAD"
       call message(AppMsg,Sparse)
    else if (noutside > 0) then
       write(errstr,*) noutside, " events outside the window"
       call message(AppMsg,Sparse)
    endif
    write(errstr,*) nbelow, " events below the EMDH lower threshold"
    call message(AppMsg,Sparse)
    write(errstr,*) nabove, " events above the EMDH upper threshold"
    call message(AppMsg,Sparse)
    if (event2 < 1) then
       write(errstr,*)  "No event survived the cleaning. ",  &
                        "The output file will be empty."
       call warning("reanalyze13")
    else
       write(errstr,*) event2, " events in output"
       call message(AppMsg,Sparse)
    endif

! Deallocate the buffer arrays

    deallocate(evtbuf, orderbuf, evtbuf2, orderbuf2)

! Set the new columns to 0 and release them

    call add_new_columns(ev_tab)

! Write the keywords summing-up the operations

    call setEventAttribute(ev_tab,EVATT_OUTSIDE_THRESHOLDS)

    if (oncutbad) then
       call keywordDone(block(ev_tab), taskname, cutbad_on)
       call setEventAttribute(ev_tab,EVATT_ON_BADPIX)
       call release(bad1_tab)
    endif

    if (onbinodal) then
       call keywordDone(block(ev_tab), taskname, "BINODAL")
       call addFilename(block(ev_tab),evfile2,"Second event file :")
       call addFilename(block(ev_tab),frfile2,"Second frame file :")
       deallocate(nvalid2)
       call release(parent(fr2_tab))
       call release(parent(ev2_tab))
    endif

    if (ondiago) then
       deallocate(diagonal)
       call keywordDone(block(ev_tab), taskname, diago_on)
       call setEventAttribute(ev_tab,EVATT_DIAGONAL)
    endif

    call setEmosRejectionMask(ev_tab)

    call message(AppMsg,Verbose,"Exiting REANALYZE")

  end subroutine reanalyze

! Remove all pixels outside the CCD and corresponding to bad pixels
! This is done (together with BINODAL) by back projecting the events
! frame by frame to a CCD map and reanalyzing it
! The possibility exists (for 5 pixel events) of creating 2 events from 1

! Should maybe reject entirely events (not only pixels) centered
! on bad pixels (as is done on-board) without reanalysing them
! to avoid creating false "echoes" of bright bad pixels

  subroutine cutBad(frame0,frame1,edumode,patterns,badx,bady,badext,  &
                    threshold,orderbuf,evtbuf,orderbuf2,evtbuf2, nbad, &
                    patterns_tim)

! frame0,frame1 : First and last frame to be analyzed
! edumode   : EDU mode (for edusoft)
!           : 0=Transparent, 1=Timing, 2=Threshold, 3=Imaging
! patterns  : Pattern library read from the CAL
! badxxx    : Bad pixel characteristics
! threshold : EDU threshold
! orderbuf  : array of good indices for evtbuf
! evtbuf    : Temporary storage for partial event list
! orderbuf2 : array of good indices for evtbuf2
! evtbuf2   : Temporary storage for partial event list for node2
! nbad      : Number of events strongly affected
! patterns_tim : pattern timing (optional)

    integer,             intent(in) :: frame0, frame1, edumode
    integer(kind=int16), intent(in) :: threshold
    integer(kind=int8), dimension(:,:,:), pointer  :: patterns
    integer(kind=int16),dimension(:), pointer      :: badx, bady, badext
    integer(kind=int32),dimension(:),intent(inout) :: orderbuf
    type(sas_evt),      dimension(:),intent(inout) :: evtbuf
    integer(kind=int32),dimension(:),intent(inout) :: orderbuf2
    type(sas_evt),      dimension(:),intent(inout) :: evtbuf2
    integer, intent(out)  :: nbad
    integer(kind=int8), dimension(:,:,:), pointer, optional  :: patterns_tim

    integer(kind=int16), parameter :: edgeflag = 1, badpixflag = 2
    logical :: badcosmic
    integer :: count0, count1, newevents, count0_n2, count1_n2
    integer :: irow,jrow,lastrow,order,jbad,xbad,ybad,x0,y0
    integer :: npattern, ir
    integer, dimension(:), allocatable :: order2
    integer(kind=int32), dimension(:,:), allocatable :: image, image2
    integer(kind=int16), dimension(:,:), allocatable :: modif
    integer(kind=int16), dimension(:),   allocatable :: offX
    integer(kind=int16), dimension(:),   allocatable :: offY

    type(edu_pattern) :: edupat(edu_npat)
    type(edu_out)     :: eduout
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1) :: patabove
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1) :: patabove_tim
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1) :: patabove2
    integer(kind=int16), dimension(edu_nsid,edu_nsid)      :: im
    logical, dimension(:), pointer :: cosmic, cosmic_img
    logical :: reject2node, ok

    call message(AppMsg,Noisy,"Entering CUT_BAD ...")

    call preparePatterns(patterns,patabove)

! Timing patterns (note there are only 4 timing patterns)

    if (timeMode)  call preparePatterns(patterns_tim,patabove_tim)

! For 2nd node, patabove2 is derived from patabove with an x-inversion

    patabove2(-1,:,:) = patabove( 1,:,:)
    patabove2( 0,:,:) = patabove( 0,:,:)
    patabove2( 1,:,:) = patabove(-1,:,:) 

    call pat_init(patterns,edupat)  ! Initialize EDUSOFT
    call cosmicrayPattern(patterns, cosmic)   ! Get non-isolated patterns

! In Timing mode cosmic is used for Timing patterns, cosmic_img for Imaging

    if (timeMode) then
       allocate(cosmic_img(0:size(cosmic)-1))
       cosmic_img = cosmic
       deallocate(cosmic)
       call cosmicrayPattern(patterns_tim, cosmic)
    endif

    allocate(image(-2:x_size_window,-2:y_size_window))
    if ( onbinodal ) &
         allocate(image2(-2:x_size_window,-2:y_size_window))
    allocate(modif(-2:x_size_window,-2:y_size_window))
    allocate(offX(-2:x_size_window),offY(-2:y_size_window))

! Widen the offset arrays

    offX(0:x_size_window-1) = offsetX(0:x_size_window-1,0)
    offX(-2:-1) = offX(0)
    offX(x_size_window) = offX(x_size_window-1)
    if (.not. timeMode) then
       offY(0:y_size_window-1) = offsetY(0:y_size_window-1,0)
       offY(-2:-1) = offY(0)
       offY(y_size_window) = offY(y_size_window-1)
    else

! In Timing mode the Y Offset is expanded from 602 to 1024 by repeating
! the last value

       offY(0:size(offsetY)-1) = offsetY(0:size(offsetY)-1,0)
       offY(size(offsetY):y_size_window-1) = offY(size(offsetY)-1)

       offY(-2:-1) = offY(0)
       offY(y_size_window) = offY(y_size_window-1)
    endif


! Array saying which events need to be reanalyzed (for optimisation)
! Border events and next to bad pixels

    if (imageMode) then
       modif = 0
       modif(-1:underx,:) = edgeflag
       modif(:,-1:undery) = edgeflag

! Binodal
 
       if( onbinodal ) then
          modif(x_size_window_bin-overx-2:x_size_window_bin,:) = edgeflag
       else
          modif(x_size_window-overx-1:x_size_window,:) = edgeflag         
       endif
       modif(:,y_size_window-overy-1:y_size_window) = edgeflag
       if ( oncutbad ) then
          do jbad = 0,size(badx)-1
             xbad = badx(jbad) + underx - 1
             ybad = bady(jbad) + undery - 1
             modif(xbad-1:xbad+1,ybad-1:ybad+badext(jbad)) =  &
                modif(xbad-1:xbad+1,ybad-1:ybad+badext(jbad)) + badpixflag
          enddo
       endif
    endif

    count0 = sum(int(nvalid(frame0-frameoff:frame1-frameoff)))
    allocate(order2(2*count0))
    lastrow = maxval(orderbuf(1:count0))
    newevents = 0
    count0 = 1
    count0_n2 = 1
    nbad = 0

    do irow = frame0-frameoff,frame1-frameoff          ! Loop over frames

       if (onbinodal .and. nvalid(irow) == 0) then

! Compute all the same count1_n2

          count1_n2 = count0_n2 + nvalid2(irow+frameoff+offframe-frameoff2)-1
          count0_n2 = count1_n2 + 1
          cycle   ! No event in that frame
       endif
       count1 = count0 + nvalid(irow) - 1

! Check if the frame corresponding to the 2nd node has a flag >= flag_reject
! (in this case the events of the 2nd node are not projected and the events of 
! the 1st node at RAWX=300 will receive the flag EVATT_CLOSE_TO_CCD_BORDER )

       if ( onbinodal ) then
          if ( frameflag2 ) then
             if (quality2(irow+frameoff+offframe-frameoff2) >=flag_reject) then
                reject2node = .true.
             else
                reject2node = .false.
             endif
          else
             reject2node = .false.
          endif
       endif

! In the case of timing mode with onevrec, rawy is in the range [0,1023].
! In the projection and then analysis by edufosft process rawy-2=y0 can reach 
! -4 for rawy=0, which is out of limit of image() which is necessarily 
! image(-2:,-2:) (see projectEventsEnergy() ). 
! So we just shift by +2 the rawy coordinate in this case 
! ( Badpixel in timing are expected to be columns, so no +2 y-translation has 
! to be done ) 
! After the analysis, -2 will be applied to restore the original rawy values.
! (note that image() was dimensionned accordingly in Y with y_size_window  

       if (timeMode) then
          do order = count0, count1
             jrow = orderbuf(order)
             evtbuf(jrow)%rawy = evtbuf(jrow)%rawy + 2
          enddo
       endif

! Set the useful area to 0 (must be done globally because events can be close)

       if (imageMode) then
          do order = count0, count1
             jrow = orderbuf(order)
             x0 = evtbuf(jrow)%rawx - 2
             y0 = evtbuf(jrow)%rawy - 2
             image(x0-1:x0+edu_nsid-2,y0-1:y0+edu_nsid-2) = 0
             if( onbinodal ) image2(x0-1:x0+edu_nsid-2,y0-1:y0+edu_nsid-2) = 0

          enddo
       else if (rimagMode .or. timeMode) then
          do order = count0, count1
             jrow = orderbuf(order)
             x0 = evtbuf(jrow)%rawx - 2
             y0 = evtbuf(jrow)%rawy - 2
             image(x0-2:x0+2,y0-2:y0+2) = 0
             if (onbinodal) image2(x0-2:x0+2,y0-2:y0+2) = 0
             if (.not. timeMode) modif(x0-1:x0+1,y0-1:y0+1) = 0
          enddo
       endif

! binodal:  for events belonging to the corresponding frame,
!  - make the x-inversion  ( rawx = 613 -rawx )
!  - fill image() and image2() as for the 1st node

       if( onbinodal ) then
          count1_n2 = count0_n2+nvalid2(irow+frameoff+offframe-frameoff2)-1

          if ( (nvalid2(irow+frameoff+offframe-frameoff2) /= 0) &
               .and. .not.reject2node  )  then

             do order = count0_n2, count1_n2
                jrow = orderbuf2(order)

                evtbuf2(jrow)%rawx = & 
                     x_size_window+overx-2 - evtbuf2(jrow)%rawx

                x0 = evtbuf2(jrow)%rawx - 2
                y0 = evtbuf2(jrow)%rawy - 2
                if (imageMode) then
                   image(x0-1:x0+edu_nsid-2,y0-1:y0+edu_nsid-2) = 0
                   image2(x0-1:x0+edu_nsid-2,y0-1:y0+edu_nsid-2) = 0
                else if (rimagMode) then
                   image(x0-2:x0+2,y0-2:y0+2) = 0
                   image2(x0-2:x0+2,y0-2:y0+2) = 0
                   modif(x0-1:x0+1,y0-1:y0+1) = 0
                endif
             enddo

          endif
       endif

! Project the events onto the image

       if (imageMode) then
          call projectEventsEnergy(evtbuf(orderbuf(count0:count1)), &
               patabove, image)
         
          if (onbinodal) then
             if (count0_n2 <= count1_n2 .and. .not.reject2node)  &
                  call projectEventsEnergy(  &
                     evtbuf2(orderbuf2(count0_n2:count1_n2)),patabove2,image2)
          endif

       else if (timeMode) then
          call projectEventsEnergy(evtbuf(orderbuf(count0:count1)), &
               patabove_tim, image)

! In Threshold mode one event is one pixel
! modif is used to store original image (before CUT_BAD)

       else if (rimagMode) then
          do order = count0, count1
             jrow = orderbuf(order)
             x0 = evtbuf(jrow)%rawx - 2          ! Center of the event
             y0 = evtbuf(jrow)%rawy - 2
             image(x0,y0) = evtbuf(jrow)%e1
             modif(x0,y0) = image(x0,y0)
          enddo
          if ( onbinodal ) then
             do order = count0_n2, count1_n2
                jrow = orderbuf2(order)
                x0 = evtbuf2(jrow)%rawx - 2          ! Center of the event
                y0 = evtbuf2(jrow)%rawy - 2
                image2(x0,y0) = evtbuf2(jrow)%e1
                modif(x0,y0) = image2(x0,y0)
             enddo
          endif
     
       endif

! Set the bright (not dead) pixels to 0
! Their coordinates are 1,600, not 0,609

       if (oncutbad) then
          do jbad = 0,size(badx)-1
             ybad = bady(jbad) + undery - 1
             image(underx-1+badx(jbad),ybad:ybad+badext(jbad)-1) = 0  
             if( onbinodal ) &
                  image2(underx-1+badx(jbad),ybad:ybad+badext(jbad)-1) = 0  
          enddo
       endif

! Set the pixels outside the CCD to 0

       image(-1:underx-1,:) = 0
       image(x_size_window-overx:x_size_window,:) = 0
       if(.not. timeMode) then
          image(:,-1:undery-1) = 0
          image(:,y_size_window-overy:y_size_window) = 0
       endif

       if( onbinodal .and. .not.reject2node ) then

! Pixel outside the CCD are set to 0 for the 2 nodes. This is done by
! combining the 2 projections. This explains why the 'Set the useful area to 0'
! has been duplicated on the 2 images. 

!          image(x_size_window_bin-overx:x_size_window,:) =  &
!               image2(x_size_window_bin-overx:x_size_window,:)
! In order to optimize it is enough to do:

          image(x_size_window_bin-overx:x_size_window_bin,:) =  &
               image2(x_size_window_bin-overx:x_size_window_bin,:)

       endif

! Run EDUSOFT locally over all events
! In theory (for BINODAL) the GATTI should be included   ml ??
! (threshold effect at high energy). Not done here.

! Only the events the central pixel of which is located on the half part of
! the image corresponding to the 1st node have to be kept. This is obtained
! by looping only on evtbuf() events.

       nvalid(irow) = 0     ! Recompute nvalid
       do order = count0, count1
          jrow = orderbuf(order)

! Non isolated patterns must not be allowed to form new events
! Those outside the CCD or incorporating bad pixels are flagged
! The same test is used when onevrec is set to false

          npattern = evtbuf(jrow)%pattern
          x0 = evtbuf(jrow)%rawx - 2          ! Center of the event
          y0 = evtbuf(jrow)%rawy - 2
          if (cosmic(npattern) .or. .not. onevrec)  then
             newevents = newevents + 1
             nvalid(irow) = nvalid(irow) + 1   
             order2(newevents) = jrow

! Do nothing in Timing mode

             if (imageMode .and. iand(modif(x0,y0),badpixflag) > 0) then
                if (x0 >= underx .and. x0 < x_size_window-overx .and.  &
                    y0 >= undery .and. y0 < y_size_window-overy) then

! Test not foolproof next to edges where event may have been clipped

                   badcosmic = (evtbuf(jrow)%e1 + evtbuf(jrow)%e2) >  &
        (image(x0,y0) + sum(patabove(:,:,npattern)*image(x0-1:x0+1,y0-1:y0+1)))
                   if (badcosmic) then
                      nbad = nbad + 1
                      evtbuf(jrow)%flag = ior(evtbuf(jrow)%flag, &
                                              EVATT_ON_BADPIX)
                   endif
                endif
             endif
          else

             if (imageMode .and. modif(x0,y0) > 0) then

! Add the offsets (in coordinates 0,609) and call EDUSOFT
! Center the event in 2,2 in the small map (we already know it is isolated)

                im = image(x0-1:x0+edu_nsid-2,y0-1:y0+edu_nsid-2)

                do ir = 1,edu_nsid
                   im(ir,:) = im(ir,:) + offX(x0+ir-2)
                   im(:,ir) = im(:,ir) + offY(y0+ir-2)
                enddo

                call edusoft(edumode,edupat,edu_nsid,edu_nsid,im,threshold,&
                     offX(x0-1:x0+edu_nsid-2),offY(y0-1:y0+edu_nsid-2),0,0,&
                     es_nmax,eduout)
     
                nvalid(irow) = nvalid(irow) + eduout%nevent

! In binodal, in the peculiar case of events :
! - where the central pixel is located at the border between the 2 nodes
! - where the pixel located beside the central pixel (at same y) has the same
!   value as the central pixel
! the event will be rejected in the 2 nodes by cleanhalf() (see reanalyse()) 
! because for the EDU the central pixel is considered as the brightest one if 
! the following conditions are true :
!    <=  <  <  
!    <=  cp <  
!    <=  <= <  
! To solve this problem, these events are detected here above and in the case
! of 1st node are shifted by -1 in x direction, so that these events will be 
! attributed to node 0.

                if( onbinodal ) then
                   if ( eduout%nevent > 0 .and. (ccdnode==0) ) then
                      do ir = 1,eduout%nevent

! Test  near border in image() coordinates (in the opposite current node side)
! and   pixel left central pixel == central pixel  
! and   node 0 (1st node)

                         if( (eduout%evt(ir)%x+x0-3==x_size_window_bin-overx) &
                             .and.  (eduout%evt(ir)%e1 == &
                     image(eduout%evt(ir)%x+x0-4,eduout%evt(ir)%y+y0-3)) ) then
                            
                            eduout%evt(ir)%x = eduout%evt(ir)%x - 1
                         endif
                      enddo
                   endif
                endif

                select case(eduout%nevent)
                case(0)
                   nbad = nbad + 1
                case(1)
                   newevents = newevents + 1
                   nbad = nbad + 1
                   if (image(x0,y0) == 0) then  ! Create new event
                      lastrow = lastrow + 1
                      order2(newevents) = lastrow
                      call copyEvent(jrow, lastrow, x0-1, y0-1,  &
                                     eduout%evt(1), evtbuf)
                   else  ! Modify old event
                      order2(newevents) = jrow
                      call copyEvent(jrow, jrow, x0-1, y0-1,  &
                                     eduout%evt(1), evtbuf)
                   endif
                case(2)
                   write(errstr,*) "two events created out of one ",  &
                      "at x0=",x0," y0=",y0," in frame ",evtbuf(jrow)%frame
                   call message(AppMsg,Verbose)

                   nbad = nbad + 2
                   do ir = 1,2
                      lastrow = lastrow + 1
                      newevents = newevents + 1
                      order2(newevents) = lastrow
                      call copyEvent(jrow, lastrow, x0-1, y0-1,  &
                                     eduout%evt(ir), evtbuf)
                   enddo
                case default
                   write(errstr,*) eduout%nevent, "events found ",  &
                                   "at x0=", x0, " y0=", y0, " in frame ",  &
                                   evtbuf(jrow)%frame, ". Contact developer"

                   call error("cutBad01")
                end select

! When rimagMode run edusoft only when several pixels were detected locally

             else if (rimagMode .and. onevrec .and.  &
                      sum(image(x0-1:x0+1,y0-1:y0+1)) > image(x0,y0)) then

! Add the offsets (in coordinates 0,609) and call EDUSOFT
! Center the event in 3,3 in the small map
! Keep only the centered output event

                im = image(x0-2:x0+2,y0-2:y0+2)
                do ir = 1,edu_nsid
                   im(ir,:) = im(ir,:) + offX(x0+ir-3)
                   im(:,ir) = im(:,ir) + offY(y0+ir-3)
                enddo
                call edusoft(edumode,edupat,edu_nsid,edu_nsid,im,threshold,  &
                            offX(x0-2:x0+2),offY(y0-2:y0+2),0,0,es_nmax,eduout)
                nbad = nbad + 1

                if( onbinodal ) then
                   if ( eduout%nevent > 0 .and. (ccdnode==0) ) then
                      do ir = 1,eduout%nevent

! Test  near border in image() coordinates (in the opposite current node side)
! and   pixel left central pixel == central pixel  
! and   node 0 (1st node)

                         if( (eduout%evt(ir)%x+x0-4==x_size_window_bin-overx) &
                             .and.  (eduout%evt(ir)%e1 == &
                      image(eduout%evt(ir)%x+x0-5,eduout%evt(ir)%y+y0-4))) then
                            
                            eduout%evt(ir)%x = eduout%evt(ir)%x - 1
                         endif
                      enddo
                   endif
                endif

                do ir = 1,eduout%nevent
                   if (eduout%evt(ir)%x == 4 .and.  &
                       eduout%evt(ir)%y == 4) then
                      nbad = nbad - 1
                      nvalid(irow) = nvalid(irow) + 1
                      lastrow = lastrow + 1
                      newevents = newevents + 1
                      order2(newevents) = lastrow
                      call copyEvent(jrow, lastrow, x0-2, y0-2,  &
                                     eduout%evt(ir), evtbuf)
                   endif
                enddo

             else if (rimagMode .and. image(x0,y0) == 0) then
                nbad = nbad + 1

! In timing EDUSOFT used if the central pixel was not set to 0 (bad pixel)

             else if (timeMode .and. image(x0,y0) > 0) then

! Add the offsets (in coordinates 0,609) and call EDUSOFT
! Center the event in 3,3 in the small map
! Keep only the centered output event

                im = image(x0-2:x0+2,y0-2:y0+2)
                do ir = 1,edu_nsid
                   im(ir,:) = im(ir,:) + offX(x0+ir-3)
                   im(:,ir) = im(:,ir) + offY(y0+ir-3)
                enddo
                call edusoft(edumode,edupat,edu_nsid,edu_nsid,im,threshold,  &
                            offX(x0-2:x0+2),offY(y0-2:y0+2),0,0,es_nmax,eduout)
    
                ok = .false.
                if( eduout%nevent > 0 ) then
                   do ir = 1,eduout%nevent
                      if (.not.cosmic_img(eduout%evt(ir)%pattern)) then
                         if ( (abs(eduout%evt(ir)%x - 4) <= 1) .and.  &
                              (abs(eduout%evt(ir)%y - 4) <= 1) ) ok = .true.
                      endif
                   enddo
                endif
                if (ok) then   ! keep the event
                   newevents = newevents + 1
                   nvalid(irow) = nvalid(irow) + 1
                   order2(newevents) = jrow
                else           ! flag it as cosmic
                   nbad = nbad + 1
                   nvalid(irow) = nvalid(irow) + 1
                   newevents = newevents + 1
                   order2(newevents) = jrow
                   evtbuf(jrow)%flag = ior(evtbuf(jrow)%flag,EVATT_COSMIC_RAY)
                endif

             else if (timeMode .and. image(x0,y0) == 0) then
                nbad = nbad + 1

             else
                newevents = newevents + 1
                nvalid(irow) = nvalid(irow) + 1
                order2(newevents) = jrow
             endif

! Keep the original bad pixel events and flag them for rejection
! In threshold mode keep them only if main pixel of an event (local maximum)
! Pixels at equal (maximum) height are all kept
! Must distinguish bad pixels from "outside CCD"

             if (image(x0,y0) == 0) then
                if (imageMode .or. (rimagMode .and. (.not. onevrec .or.  &
                   maxval(modif(x0-1:x0+1,y0-1:y0+1)) == modif(x0,y0)))) then
                   nvalid(irow) = nvalid(irow) + 1
                   newevents = newevents + 1
                   order2(newevents) = jrow
                   if (x0 >= underx .and. x0 < x_size_window-overx .and.  &
                       y0 >= undery .and. y0 < y_size_window-overy)       &
                      evtbuf(jrow)%flag = ior(evtbuf(jrow)%flag,     &
                                              EVATT_ON_BADPIX)
                else if (timeMode) then
                   nvalid(irow) = nvalid(irow) + 1
                   newevents = newevents + 1
                   order2(newevents) = jrow
                   if (x0 >= underx .and. x0 < x_size_window-overx ) &
                        evtbuf(jrow)%flag = ior(evtbuf(jrow)%flag, &
                        EVATT_ON_BADPIX )  ! No check on Y
                endif
             endif

          endif
       enddo

! Restore the original rawy (see above what was done before the projection)

       if (timeMode) then
          do order = count0, count1
             jrow = orderbuf(order)
             evtbuf(jrow)%rawy = evtbuf(jrow)%rawy - 2
          enddo
       endif

! In binodal, if the frame was flagged as rejected set events at the border 
! (ie at RAWX=306 ) with flag close_to_ccd_border

       if ( onbinodal .and. reject2node ) then
          do order = count0, count1
             jrow = orderbuf(order)
             if ( evtbuf(jrow)%rawx == x_size_window_bin-overx+1) then
                evtbuf(jrow)%flag = &
                     ior(evtbuf(jrow)%flag,EVATT_CLOSE_TO_CCD_BORDER)
             endif
          enddo
       endif

! Counters incrementation

       count0 = count1 + 1
       if ( onbinodal ) count0_n2 = count1_n2 + 1

    enddo

    orderbuf(1:newevents) = order2(1:newevents)
    deallocate(order2,cosmic)
    if (timeMode)  deallocate(cosmic_img)
    deallocate(image,modif,offX,offY)
    if ( onbinodal )  deallocate(image2)

  end subroutine cutBad

! Split diagonal events into two separate events (except Si K line)

  subroutine diago(nevents2, patterns, diagonal, emini, emaxi, offpattern,  &
                   orderbuf, evtbuf, ndiago)

! nevents2  : Number of events
! patterns  : Pattern library read from the CAL
! diagonal  : true for diagonal patterns
! emini     : Lower bound of Si line interval (ADU)
! emaxi     : Upper bound of Si line interval (ADU)
! offpattern: Difference in pattern number between diagonals and bipixels
! orderbuf  : array of good indices for evtbuf
! evtbuf    : Temporary storage for partial event list
! ndiago    : Number of diagonal events found

    integer, intent(inout)  :: nevents2
    integer(kind=int8), dimension(:,:,:), pointer  :: patterns
    logical,            dimension(:),     pointer  :: diagonal
    integer(kind=int32),dimension(:),intent(inout) :: orderbuf
    type(sas_evt),      dimension(:),intent(inout) :: evtbuf
    integer, intent(in)  :: emini, emaxi, offpattern
    integer, intent(out) :: ndiago

    integer :: lastrow, irow, jrow, frow, nevents3, patcur, ix, iy, e2cur
    integer, dimension(2*nevents2) :: order2

    call message(AppMsg,Noisy,"Entering DIAGO ...")

    lastrow = maxval(orderbuf(1:nevents2))

    nevents3 = 0
    ndiago = 0
    do jrow = 1,nevents2
       irow = orderbuf(jrow)
       nevents3 = nevents3 + 1
       order2(nevents3) = irow
       patcur = evtbuf(irow)%pattern
       e2cur = evtbuf(irow)%e2

! Jump non diagonal event

       if (.not. diagonal(patcur))  cycle

! Si K line: set PATTERN to bipixel and flag

       if (onsifluor .and.  &
            min((evtbuf(irow)%e1-emini) * (evtbuf(irow)%e1-emaxi),  &
            (e2cur-emini) * (e2cur-emaxi)) < 0) then
          evtbuf(irow)%pattern = evtbuf(irow)%pattern - offpattern
          evtbuf(irow)%flag = ior(evtbuf(irow)%flag,EVATT_DIAGONAL)
       else

! Modify initial event and add additional event at the end of the array

          lastrow = lastrow + 1
          nevents3 = nevents3 + 1
          order2(nevents3) = lastrow
          frow = evtbuf(irow)%frame - frameoff
          nvalid(frow) = nvalid(frow) + 1
          ndiago = ndiago + 1
 
          evtbuf(irow)%pattern = 0
          evtbuf(irow)%e2 = 0
          evtbuf(irow)%flag = ior(evtbuf(irow)%flag,EVATT_DIAGONAL)

          evtbuf(lastrow) = evtbuf(irow)

          evtbuf(lastrow)%e1 = e2cur
          do ix = 1,3,2
             do iy = 1,3,2
                if (patterns(patcur,iy,ix) == 1) then
                   evtbuf(lastrow)%rawx = evtbuf(irow)%rawx + ix - 2
                   evtbuf(lastrow)%rawy = evtbuf(irow)%rawy + iy - 2
                endif
             enddo
          enddo
       endif
    enddo
    nevents2 = nevents3
    orderbuf(1:nevents2) = order2(1:nevents2)

  end subroutine diago


! remove events which are located in the opposite half of the CCD

  subroutine cleanhalf( nevents2, orderbuf, evtbuf )

! nevents2  : Number of events
! orderbuf  : array of good indices for evtbuf
! evtbuf    : Temporary storage for partial event list

    integer, intent(inout)  :: nevents2
    integer(kind=int32),dimension(:),intent(inout) :: orderbuf
    type(sas_evt),      dimension(:),intent(inout) :: evtbuf

    integer :: irow, jrow, frow, nevents3

    call message(AppMsg,Noisy,"Entering CLEANHALF ...")

    nevents3 = 0
 
    do jrow = 1,nevents2
       irow = orderbuf(jrow)

       if ( evtbuf(irow)%rawx < x_size_window_bin-overx+2 ) then

          nevents3 = nevents3 + 1
          orderbuf(nevents3) = irow

       else

          frow = evtbuf(irow)%frame - frameoff
          nvalid(frow) = nvalid(frow) - 1

       endif

    enddo
    nevents2 = nevents3
 

  end subroutine cleanhalf


! Flag events corresponding to flagged frames
! Flag cosmic-rays

  subroutine cutGTI(ev_tab, fr_tab)

! ev_tab   : Handle to the event table
! fr_tab   : Handle to the frame table

    type(TableT), intent(in)   :: ev_tab, fr_tab

    integer :: irow,jrow,count,nflag,ncosmics,ip
    integer, dimension(-1:31) :: cosmic
    character(len=8)  :: keyword

    call message(AppMsg,Verbose,"Entering CUT_GTI ...")

    patt_col = column(ev_tab, patt, Read)
    pattern => int8Data(patt_col)

! Flag events in bad frames

    count = 0
    nflag = 0
    do irow = firstframe-frameoff,lastframe-frameoff        ! Loop over frames
       if (nvalid(irow) == 0) cycle   ! No event in that frame
       if (quality(irow) >= flag_reject) then
          do jrow = count,count+nvalid(irow)-1
             qualev(jrow) = ior(qualev(jrow),EVATT_IN_BAD_FRAME)
          enddo
          nflag = nflag + nvalid(irow)
       else if (quality(irow) > 0) then
          do jrow = count,count+nvalid(irow)-1
             qualev(jrow) = ior(qualev(jrow),EVATT_IN_SPOILED_FRAME)
          enddo
       endif
       count = count + nvalid(irow)
    enddo
    write(errstr,*) nflag, " events outside good frames"
    call message(AppMsg,Sparse)
    call setEventAttribute(ev_tab,EVATT_IN_BAD_FRAME)
    call setEventAttribute(ev_tab,EVATT_IN_SPOILED_FRAME)

! Flag "cosmic-ray" events

    cosmic = 0
    do ip = 1,32
       write(keyword,*) ip
       keyword = adjustl(keyword)
       keyword = "CRPAT" // keyword(1:3)
       if (.not. hasAttribute(fr_tab, keyword))  exit
       count = int16Attribute(fr_tab,keyword)
       cosmic(count) = 1
       call setAttribute(ev_tab,attribute(fr_tab,keyword))
       write(errstr,*) count, " is declared as a cosmic-ray pattern."
       call message(AppMsg,Verbose)
    enddo
    if (hasAttribute(fr_tab,"COSMICSI"))  &
       call setAttribute(ev_tab,attribute(fr_tab,"COSMICSI"))
    if (hasAttribute(fr_tab,"COSMICOU"))  &
       call setAttribute(ev_tab,attribute(fr_tab,"COSMICOU"))

    ncosmics = 0
    do jrow = 0,size(pattern)-1
       count = pattern(jrow)
       if (cosmic(count) > 0) then
          qualev(jrow) = ior(qualev(jrow),EVATT_COSMIC_RAY)
          ncosmics = ncosmics + 1
       endif
    enddo
    write(errstr,*) ncosmics, " cosmic-ray events"
    call message(AppMsg,Sparse)
    call setEventAttribute(ev_tab,EVATT_COSMIC_RAY)

    call release(patt_col)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, cutgti_on)

  end subroutine cutGTI

! Transform raw coordinates to standard node coordinates
! Subtract 2 because of EDU convention to send coords of upper-right corner
! Subtract underscan to get back to the true CCD area
! Add 1 to have coordinates starting at 1 (not 0)

  subroutine putXY(ev_tab)

! ev_tab   : Handle to the event table

    type(TableT), intent(in)  :: ev_tab

    integer :: windX0,windY0,windDX,windDY,windX1,windY1
    integer :: numb

    if (wasDone(block(ev_tab),changexy)) then
       call warning("putXY10","PUT_XY was already applied. Not redone.")
       return
    else
       call message(AppMsg,Verbose,"Entering PUT_XY ...")
    endif

! Flag events outside the window
! Adjust window parameters. In full frame mode remove under/overscans
! In window mode remove 2 (EDU cannot find events in last 2 lines/columns)
! Remove side columns with offsets set to 4095
! Write TLMIN/TLMAX keywords

! Should check that it works in binodal mode

! Operations on X axis

    windX0 = int32Attribute(ev_tab,"WINDOWX0")
    windDX = int32Attribute(ev_tab,"WINDOWDX")
    write(errstr,*) " Input: WINDOWX0 =",windX0,"     WINDOWDX =",windDX
    call message(AppMsg,Verbose)

! Take out the under and overscans

    windX1 = min(windX0+windDX-2,x_size_window-overx) - 1
    windX0 = max(windX0,underx)

! Take out the columns with offsets set to 4095

    do while (offsetX(windX0,0) == 4095)
       windX0 = windX0 + 1
    enddo
    do while (offsetX(windX1,0) == 4095)
       windX1 = windX1 - 1
    enddo
    windDX = windX1 - windX0 + 1

! Switch to PIXCOORD1 coordinates

    windX0 = windX0 - underx + 1
    windX1 = windX1 - underx + 1
    write(errstr,*) "Output: WINDOWX0 =",windX0,"     WINDOWDX =",windDX
    call message(AppMsg,Verbose)
    call setAttribute(ev_tab, "WINDOWX0", windX0,  &
                      comment=label(attribute(ev_tab,"WINDOWX0")))
    call setAttribute(ev_tab, "WINDOWDX", windDX,  &
                      comment=label(attribute(ev_tab,"WINDOWDX")))
    call setAttribute(rawx_col, "TLMIN", 1-underx,  &
                      comment="Legal minimum for "//rawx)
    call setAttribute(rawx_col, "TLMAX", x_size_window-underx,  &
                      comment="Legal maximum for "//rawx)
    call setAttribute(rawx_col, "TDMIN", windX0,  &
                      comment="Actual minimum for "//rawx)
    call setAttribute(rawx_col, "TDMAX", windX1,  &
                      comment="Actual maximum for "//rawx)

    if (ctimeMode) then     ! Special case of COMPRESSED TIMING data
       windX0 = windX0 - (1 - 2 - underx)
       windX1 = windX1 - (1 - 2 - underx)
       where (posx /= -1)
          where (posx < windX0 .or. posx > windX1)
             qualev = ior(qualev,EVATT_OUT_OF_CCD_WINDOW)
          endwhere
          posx = posx + (1 - 2 - underx)
       endwhere
    else
       posx = posx + (1 - 2 - underx)
       where (posx < windX0 .or. posx > windX1)
          qualev = ior(qualev,EVATT_OUT_OF_CCD_WINDOW)
       endwhere
    endif

! Operations on Y axis (not in TIMING mode)

    if (imageMode .or. rimagMode) then
       posy = posy + (1 - 2 - undery)

       windY0 = int32Attribute(ev_tab,"WINDOWY0")
       windDY = int32Attribute(ev_tab,"WINDOWDY")
       write(errstr,*) " Input: WINDOWY0 =",windY0,"     WINDOWDY =",windDY
       call message(AppMsg,Verbose)
       windY1 = min(windY0+windDY-2,y_size_window-overy) - 1
       windY0 = max(windY0,undery)
       do while (offsetY(windY0,0) == 4095)
          windY0 = windY0 + 1
       enddo
       do while (offsetY(windY1,0) == 4095)
          windY1 = windY1 - 1
       enddo
       windDY = windY1 - windY0 + 1
       windY0 = windY0 - undery + 1
       windY1 = windY1 - undery + 1
       write(errstr,*) "Output: WINDOWY0 =",windY0,"     WINDOWDY =",windDY
       call message(AppMsg,Verbose)
       call setAttribute(ev_tab, "WINDOWY0", windY0,  &
                         comment=label(attribute(ev_tab,"WINDOWY0")))
       call setAttribute(ev_tab, "WINDOWDY", windDY,  &
                         comment=label(attribute(ev_tab,"WINDOWDY")))
       call setAttribute(rawy_col, "TLMIN", 1-undery,  &
                         comment="Legal minimum for "//rawy)
       call setAttribute(rawy_col, "TLMAX", y_size_window-undery,  &
                         comment="Legal maximum for "//rawy)
       call setAttribute(rawy_col, "TDMIN", windY0,  &
                         comment="Actual minimum for "//rawy)
       call setAttribute(rawy_col, "TDMAX", windY1,  &
                         comment="Actual maximum for "//rawy)

       where (posy < windY0 .or. posy > windY1)
          qualev = ior(qualev,EVATT_OUT_OF_CCD_WINDOW)
       endwhere
    else   ! Timing mode
       call setAttribute(rawy_col, "TLMIN", 0, comment="Minimum for "//rawy)
       call setAttribute(rawy_col, "TLMAX", 1023, comment="Maximum for "//rawy)
    endif

    numb = sum(iand(qualev,EVATT_OUT_OF_CCD_WINDOW)/EVATT_OUT_OF_CCD_WINDOW)
    write(errstr,*) numb, " events outside window"
    call message(AppMsg,Sparse)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, changexy)
    call setEventAttribute(ev_tab,EVATT_OUT_OF_CCD_WINDOW)

  end subroutine putXY

! Compute DETX and DETY along the Z and Y satellite axes

  subroutine putDetCoords(ev_tab)

! ev_tab    : Handle to the input/output event table

    type(TableT), intent(in)   :: ev_tab

    type(ColumnT)  :: detx_col, dety_col
    integer(kind=int16), dimension(:), pointer   :: xdet, ydet, dumx, dumy
    real(double) :: pi, conv, step, focalLength
    integer(kind=int16) :: xmin, xmax, ymin, ymax
    integer :: nrows, xraw, yraw

    call message(AppMsg,Verbose,"Entering PUT_DET_COORDS")

    call CAL_getMiscellaneousData('FOCAL_LENGTH',focalLength)
    write(errstr,*) "Focal length: ", focalLength
    call message(AppMsg,Verbose)
    pi = acos(0.d0) * 2.d0
    step = 3600.d0 / det_unit
    conv = step * 180.d0 / pi

    detx_col = column(ev_tab, detx, Modify)
    dety_col = column(ev_tab, dety, Modify)

    nrows = numberOfRows(ev_tab)
    if (nrows > 0) then
       xdet => int16Data(detx_col)
       ydet => int16Data(dety_col)

! Timing modes: Take RAWY (and RAWY if not in data) from keywords.
!               (coherent with time definitions in emframes)
!               Could use posx+0.5 as position for bipixel events

       if (timeMode .or. ctimeMode) then
          if (hasAttribute(ev_tab,"SRC_RAWX")) then
             xraw = int32Attribute(ev_tab,"SRC_RAWX")
          else
             xraw = EMOS_MAX_X / 2
          endif
          if (hasAttribute(ev_tab,"SRC_RAWY")) then
             yraw = int32Attribute(ev_tab,"SRC_RAWY")
          else
             yraw = EMOS_MAX_Y / 2
          endif
       endif

       allocate(dumx(0:nrows-1),dumy(0:nrows-1))
       dumx = posx
       if (ctimeMode) then
          where (posx == -1)
             dumx = xraw
          endwhere
       endif
       if (imageMode .or. rimagMode) then
          dumy = posy
       else
          dumy = yraw
       endif

! CAL_rawXY2mm includes randomisation

       call CAL_setState(randomize=xyrandom)
       call getDetCoords(dumx, dumy, xdet, ydet, conv, focalLength, .true.)
       deallocate(dumx,dumy)
    endif

! Get extreme chip coordinates

    call CAL_setState(randomize=.false.)
    call getDetBorders("TLMIN","TLMAX",conv,focalLength,xmin,xmax,ymin,ymax)

! Write keyword stating the system which was used

    call setAttribute(ev_tab, "DET_SYST", detSysType,  &
                      comment="System for camera coordinates")
    if (xyrandom) then
       call setAttribute(ev_tab, "RAND_XY", 1,  &
                         comment="Camera and sky coordinates are randomized")
    else
       call setAttribute(ev_tab, "RAND_XY", 0,  &
                         comment="Spatial coordinates are not randomized")
    endif

! Write TLMIN/MAX and WCS keywords for detector coordinates
! The first part of TCTYP should not be RA, DEC

    call setAttribute(detx_col,"TCRPX",det_offset,comment="Center for "//detx)
!    call setAttribute(detx_col,"TCTYP","RA---TAN",  &
!       comment="Projection type for "//detx)
    call setAttribute(detx_col,"TCUNI",angl_unit,  &
       comment="Physical unit for "//detx)
    call setAttribute(detx_col,"TCDLT",1./step,"degrees",  &
       comment="Pixel size for "//detx)
    call setAttribute(detx_col,"TLMIN",xmin,comment="Legal minimum for "//detx)
    call setAttribute(detx_col,"TLMAX",xmax,comment="Legal maximum for "//detx)

    call setAttribute(dety_col,"TCRPX",det_offset,comment="Center for "//dety)
!    call setAttribute(dety_col,"TCTYP","DEC--TAN",  &
!       comment="Projection type for "//dety)
    call setAttribute(dety_col,"TCUNI",angl_unit,  &
       comment="Physical unit for "//dety)
    call setAttribute(dety_col,"TCDLT",1./step,"degrees",  &
       comment="Pixel size for "//dety)
    call setAttribute(dety_col,"TLMIN",ymin,comment="Legal minimum for "//dety)
    call setAttribute(dety_col,"TLMAX",ymax,comment="Legal maximum for "//dety)

    if (hasAttribute(rawx_col,"TDMIN")) then
       call getDetBorders("TDMIN","TDMAX",conv,focalLength,xmin,xmax,ymin,ymax)
       call setAttribute(detx_col, "TDMIN", xmin,  &
                         comment="Actual minimum for "//detx)
       call setAttribute(detx_col, "TDMAX", xmax,  &
                         comment="Actual maximum for "//detx)
       call setAttribute(dety_col, "TDMIN", ymin,  &
                         comment="Actual minimum for "//dety)
       call setAttribute(dety_col, "TDMAX", ymax,  &
                         comment="Actual maximum for "//dety)
    endif

    call release(detx_col)
    call release(dety_col)

    call message(AppMsg,Verbose,"Exiting PUT_DET_COORDS")

  end subroutine putDetCoords

! Assign time to events and create the exposure extension

  subroutine putTime(ev_tab)

! ev_tab   : Handle to the event table

    type(TableT), intent(in)   :: ev_tab

    type(DataSetT) :: ev_set
    type(TableT)   :: ex_tab
    type(ColumnT)  :: tiev_col, tiex_col, tdel_col, frac_col
    integer(kind=int16), dimension(:), pointer :: zero
    real(kind=single), dimension(:), pointer :: duration, fraction, fluct
    real(kind=double), dimension(:), pointer :: timeex, timeev
    type(ModeParameterDataType)  :: modeParameters

    integer, parameter :: nrowsTiming = 1024
    integer(int16) :: testflag
    character(len=80) :: comm
    integer :: irow, jrow, count, ievt, f1, f2
    real(kind=double) :: dtime, transfertime

    call message(AppMsg,Verbose,"Entering PUT_TIME ...")

! Define EXPOSURE extension
! If the extension already exists, it may not have the right length
! Better restart from scratch

    ev_set = parent(ev_tab)
    if (hasBlock(ev_set, expos_ext))  call deleteBlock(ev_set, expos_ext)

    call CAL_getModeParameters(modeParameters)

! Transfer time is not counted as dead time in Timing mode
! Times are in ms in the CCF

    if (timeMode .or. ctimeMode) then
       transfertime = 0.d0
    else
       transfertime = modeParameters%shiftTime * 1.e-3
    endif
    
! Nothing can be done if no frame file

    if (frameread) then

! Fill EXPOSURE extension
! TIME must be the center of the bin as in OGIP/93-003

       ex_tab   = addTable (ev_set, expos_ext, lastframe-firstframe+1,  &
                            "time steps and efficiency")

       tiex_col = addColumn(ex_tab, time,     Real64, time_unit,  &
                            comment="time of the center of each frame")
       tdel_col = addColumn(ex_tab,"TIMEDEL", Real32, time_unit,  &
                         comment="duration of each frame")
       frac_col = addColumn(ex_tab,"FRACEXP", Real32, "FRACTION",  &
                         comment="effective time / real time")

       timeex   => real64Data(tiex_col)
       duration => real32Data(tdel_col)
       fraction => real32Data(frac_col)

       f1 = firstframe - frameoff
       f2 = lastframe - frameoff

! In all imaging modes the time in the frame file marks the end
! of the frame transfer to the framestore area.
! The frame transfer is counted out from the frame time here

       if (modeParameters%dutyCycle < 0.9) then

! Special case for Refreshed Frame Store

          duration = modeParameters%integrationTime * 1.e-3
       else
          duration = timetag(f1:f2) - timetag(f1-1:f2-1) - transfertime

! Some (dummy) frames have identical times. Should not have negative duration.

          duration = max(duration,0.) 
       endif
       timeex = timetag(f1:f2) - transfertime - 0.5d0*duration
       if (onexpos) then
          fraction = 1. - crratio(f1:f2)
          comm = "FRACEXP accounts for the fraction lost to cosmic-rays"
          call addComment(ex_tab, comm)
       else
          fraction = 1.0
       endif

! Set FRACEXP to 0 rather than 1 in bad frames

       if (frameflag) then

!  Construct bit mask for removing HK GTI
!  Use intermediate variable to work around F95 bug on DEC with constants

          testflag = flag_reject - 1 + flag_hkgti
          testflag = not(testflag)
          write(errstr,'("Bit mask =",B17.16)') testflag
          call message(AppMsg,Verbose)
          where (iand(quality(f1:f2),testflag) > 0)
             fraction = 0.0
          endwhere
       endif

! Set event times

       if (hasColumn(ev_tab, trim(time))) then
          tiev_col = column(ev_tab, time, Modify)
       else
          tiev_col = addColumn(ev_tab, time, Real64, time_unit,  &
                               comment="time of the center of the frame")
       endif
       timeev => real64Data(tiev_col)

       if (timerandom) then
          allocate(zero(0:size(timeev)-1))
          zero = 0
          call CAL_setState(randomize=timerandom)
          call CAL_integerToReal(zero,fluct)
       endif

       count = 0
       do jrow = 0,lastframe-firstframe         ! Row in EXPOSURE extension
          irow = jrow + firstframe - frameoff   ! Row in FRAME extension
          if (nvalid(irow) == 0)  cycle   ! No event in that frame

! Timing mode:  time is computed from RAWY

          if (timeMode .or. ctimeMode) then
             dtime = duration(jrow) / nrowsTiming
             do ievt = count,count+nvalid(irow)-1
                if (posy(ievt) >= 0 .and. posy(ievt) < 1024) then
                   timeev(ievt) = timeex(jrow) + dtime *  &
                                  (posy(ievt)-0.5*(nrowsTiming-1))
                else
                   qualev(ievt) = ior(qualev(ievt),EVATT_OUT_OF_CCD_WINDOW)
                endif
             enddo

! Imaging mode: optional randomisation

          else if (timerandom) then
             timeev(count:count+nvalid(irow)-1) = timeex(jrow)  &
                       + fluct(count:count+nvalid(irow)-1) * duration(jrow)

! Imaging mode: set time to average time in frame/cycle

          else
             timeev(count:count+nvalid(irow)-1) = timeex(jrow)
          endif
          count = count + nvalid(irow)
       enddo 

       if (timerandom) then
          deallocate(zero)
          call CAL_releaseMemory(fluct)
       endif
       call release(tiev_col)

! Write keywords to sum up action

       call keywordDone(block(ev_tab), taskname, puttime_on)
       if (timeMode .or. ctimeMode)  &
          call setEventAttribute(ev_tab,EVATT_OUT_OF_CCD_WINDOW)
       if (timerandom) then
          call setAttribute(ev_tab, "RAND_TIM", 1,  &
                            comment="Time values are randomized")
       else
          call setAttribute(ev_tab, "RAND_TIM", 0,  &
                            comment="Time values are not randomized")
    endif

    else
       ex_tab   = addTable (ev_set, expos_ext, 0, "CCD-specific keywords")
    endif

    call setAttributes(attributable(ex_tab), attributable(ev_tab))
    call release(ex_tab)

  end subroutine putTime

! Analyse number of events per row and per frame

  subroutine rejectRows(ev_tab)

! ev_tab    : Handle to the event table

    type(TableT), intent(in)     :: ev_tab

    integer,           dimension(0:1023) :: yprofile
    real(kind=double), dimension(0:1023) :: yaver
    integer :: numb, irow, jrow, cnum, flagout, ngoodframes, ycur

    if (.not. wasDone(block(ev_tab),changexy)) then
       write(errstr,*) "REJECT_ROWS needs the corrected CCD coordinates. ",  &
                       "Please run emevents with ",changexy,"=Y. Not applied."
       call warning("rejectRows11")
       return
    endif

    call message(AppMsg,Verbose,"Entering REJECT_ROWS ...")

! Exclude events outside the CCD, in bad frames and cosmic-rays from stats

    flagout = EVATT_OUT_OF_CCD_WINDOW + EVATT_COSMIC_RAY + EVATT_IN_BAD_FRAME

    yaver = 0
    if (frameflag) then
       ngoodframes = count(quality < flag_reject)
    else
       ngoodframes = lastframe - firstframe + 1
    endif

! Security to avoid removing bright sources
! Add N + 2 SQRT(N) for all rows

    if (ngoodframes > 10) then   ! Must have many frames for the average
       do jrow = 0, size(posy)-1
          if (iand(qualev(jrow),flagout) == 0) then
             ycur = posy(jrow)
             yaver(ycur) = yaver(ycur) + 1
          endif
       enddo
       yaver = yaver / ngoodframes

! In Timing mode all rows are identical ==> average

       if (timeMode .or. ctimeMode)  yaver = sum(yaver) / 1024

       write(errstr,*) 'maximum number of expected events per row',  &
                       maxval(yaver)
       call message(AppMsg,Sparse)
       yaver = yaver + 2 * sqrt(yaver)
    endif

    numb = 0
    cnum = 0
    do irow = firstframe-frameoff,lastframe-frameoff
       if (nvalid(irow) == 0)  cycle   ! No event in that frame

! Build the Y profile

       yprofile = 0
       do jrow = cnum,cnum+nvalid(irow)-1
          if (iand(qualev(jrow),flagout) == 0) then
             ycur = posy(jrow)
             yprofile(ycur) = yprofile(ycur) + 1
          endif
       enddo

! Find out the bad rows

       do jrow = cnum,cnum+nvalid(irow)-1
          if (iand(qualev(jrow),flagout) == 0) then
             ycur = posy(jrow)
             if (yprofile(ycur) > maxevrow+yaver(ycur)) then
                qualev(jrow) = ior(qualev(jrow),EVATT_ON_BADROW)
                numb = numb + 1
             endif
          endif
       enddo
       cnum = cnum + nvalid(irow)
    enddo

    write(errstr,*) numb, ' events in rows with more than', maxevrow,  &
                    ' events above expected.'
    call message(AppMsg,Sparse)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, badrows_on)
    call setEventAttribute(ev_tab,EVATT_ON_BADROW)

  end subroutine rejectRows

! Analyse charge below threshold (E3)

  subroutine rejectE3(ev_tab)

! ev_tab    : Handle to the event table

    type(TableT), intent(in)     :: ev_tab

    type(ColumnT)  :: e3_col
    integer(kind=int16), dimension(:), pointer :: ener3
    integer, dimension(:), pointer :: offX, offY
    integer :: numb, irow, xori, yori
    logical :: newcoords

    call message(AppMsg,Verbose,"Entering REJECT_E3 ...")

    e3_col = column(ev_tab, ene3, Read)
    ener3 => int16Data(e3_col)
    patt_col = column(ev_tab, patt, Read)
    pattern => int8Data(patt_col)

! Do not reject bad E3 due to lines/columns with offsets set to 4095
! 10 means next to bad line/column

    allocate(offX(-2:x_size_window+1),offY(-2:y_size_window+1))
    offX = 0
    do irow = 0, x_size_window-1
       if (offsetx(irow,0) > 4000) then
          offX(irow-1:irow+1) = max(offX(irow-1:irow+1),10)
       endif
    enddo
    offY = 0
    do irow = 0, y_size_window-1
       if (offsety(irow,0) > 4000) then
          offY(irow-1:irow+1) = max(offY(irow-1:irow+1),10)
       endif
    enddo

    newcoords = wasDone(block(ev_tab),changexy)

    do irow = 0, size(ener3)-1
       if (newcoords) then
          if (iand(qualev(irow),EVATT_OUT_OF_CCD_WINDOW) > 0)  cycle
          xori = posx(irow) + underx - 1
          yori = posy(irow) + undery - 1
       else
          xori = max(min(posx(irow)-2,x_size_window-1),0)
          yori = max(min(posy(irow)-2,y_size_window-1),0)
       endif
       if (offX(xori) + offY(yori) >= 10)  cycle
       if (pattern(irow) <= 4 .and. ener3(irow) < e3min)  &
          qualev(irow) = ior(qualev(irow),EVATT_UNDERSHOOT)
    enddo

    numb = sum(iand(qualev,EVATT_UNDERSHOOT)/EVATT_UNDERSHOOT)
    write(errstr,*) numb, ' events with E3 <', e3min
    call message(AppMsg,Sparse)

    deallocate(offX, offY)
    call release(e3_col)
    call release(patt_col)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, bade3_on)
    call setEventAttribute(ev_tab,EVATT_UNDERSHOOT)

  end subroutine rejectE3

! Analyse GATTI correction

  subroutine spGatti(ev_tab, gattifile)

! ev_tab    : Handle to the event table
! gattifile : Name of the output file

    type(TableT), intent(in)     :: ev_tab
    character(len=*), intent(in) :: gattifile

    type(DataSetT) :: gat_set
    type(TableT)   :: gat_tab
    type(ColumnT)  :: spe1p_col, spe1g_col
    real(kind=single), dimension(:), pointer :: phaval
    integer, dimension(:), pointer :: spece1p
    real,    dimension(:), pointer :: spece1g

    logical :: newcoords
    integer, dimension(0:maxchan) :: spece1, specgat
    integer :: enddir, numb, count, irow, jrow, xori, yori, yoffset
    integer :: decal, e1p, chan, ch1, gat2, maxdecal, testqual
    character(len=80) :: comm, shortname

    call message(AppMsg,Verbose,"Entering SP_GATTI ...")

    if (imageMode .or. rimagMode) then
       e1_col = column(ev_tab, ene1, Modify)
       ener1 => int16Data(e1_col)
    else

! Energy is stored as PHA (real) in Timing mode
! PATTERN is required because only singles may be flagged

       e1_col = column(ev_tab, pha, Modify)
       phaval => real32Data(e1_col)
       allocate(ener1(0:size(phaval)-1))
       ener1 = int(phaval, kind=int16)
       patt_col = column(ev_tab, patt, Modify)
       pattern => int8Data(patt_col)
    endif

    if (outgatti) then
       gat_set = dataSet(gattifile, Create)
       gat_tab = addTable(gat_set, "SPECTRUM", maxchan+1,  &
                          "spectral response without the GATTI correction")
       call setAttributes(attributable(gat_tab), attributable(ev_tab))
       comm = "The abscissa is channel number"
       call addComment(gat_tab, comm)
       shortname = name(parent(ev_tab))
       enddir = index(shortname,'/',.true.)
       comm = "Input event file " &
                 // shortname(enddir+1:len_trim(shortname))
       call addHistory(gat_tab, comm)
       spe1p_col = addColumn(gat_tab, "ORIGINAL", Integer32, ener_unit,  &
                     comment="histogram of E1 values in output of the ADC")
       spe1g_col = addColumn(gat_tab, "SMOOTHED", Real32,    ener_unit,  &
                     comment="reference histogram to compare to")
       spece1p => int32Data(spe1p_col)
       spece1g => real32Data(spe1g_col)
    else
       allocate(spece1p(0:maxchan))
    endif

    newcoords = wasDone(block(ev_tab),changexy)
    spece1  = 0
    spece1p = 0
    specgat = 0
    maxdecal = 0
    testqual = EVATT_EMOS_REJECTION_MASK
    if (iand(testqual,EVATT_OUT_OF_FOV) > 0)  testqual = testqual  &
                                                       - EVATT_OUT_OF_FOV

    numb  = 0
    count = 0
    yoffset = 0     ! For Timing mode
    yori = 0        ! For Timing mode
    do irow = firstframe-frameoff,lastframe-frameoff
       if (nvalid(irow) == 0)  cycle   ! No event in that frame
       if (iand(qualev(count),EVATT_IN_BAD_FRAME) > 0) then
          count = count + nvalid(irow)
          cycle                        ! Do not analyse bad frames
       endif
       do jrow = count,count+nvalid(irow)-1

! Get the energy as seen by the ADC

          if (newcoords) then

! Security to avoid wrong indices in offsetX/Y

             if (iand(qualev(jrow),EVATT_OUT_OF_CCD_WINDOW) > 0)  cycle
             xori = posx(jrow) + underx - 1
          else
             xori = max(min(posx(jrow)-2,x_size_window-1),0)
          endif
          if (imageMode .or. rimagMode) then
             if (newcoords) then
                yori = posy(jrow) + undery - 1
             else
                yori = max(min(posy(jrow)-2,y_size_window-1),0)
             endif
             yoffset = offsetY(yori,0)
          else if (pattern(jrow) > 0) then
             cycle     ! Do not flag multiples in Timing mode
          endif
          decal = gattival(irow) + offsetX(xori,0) + yoffset
          if (decal < 0) then
             qualev(jrow) = ior(qualev(jrow),EVATT_REJECTED_BY_GATTI)
             write(errstr,*) "GATTI+offsetX+offsetY negative at row",  &
                   jrow+1, ". X (0-609) = ", xori, ". Y (0-601) = ", yori,  &
                   ". GATTI = ", gattival(irow), ". OffsetX = ",  &
                   offsetX(xori,0), ". OffsetY = ", yoffset,  &
                   ". One of these data must be wrong"
             call warning("spGatti12")
          endif

          if (ener1(jrow) < 0) then
             qualev(jrow) = ior(qualev(jrow),EVATT_REJECTED_BY_GATTI)
             write(errstr,*) "Energy negative at row", jrow+1, ". E1 = ",  &
                             ener1(jrow), ". This data must be wrong"
             call warning("spGatti13")
          endif

          e1p = ener1(jrow) + decal
          if (e1p >= maxchan) then
             qualev(jrow) = ior(qualev(jrow),EVATT_REJECTED_BY_GATTI)
             if (e1p > maxchan) then
                write(errstr,*) "reconstructed energy larger than ",  &
                      maxchan, " at row ", jrow+1, ". E1 = ",  &
                      ener1(jrow),". GATTI = ", gattival(irow),  &
                      ". X (0-609) = ", xori, ". Y (0-601) = ", yori,  &
                      ". OffsetX = ", offsetX(xori,0), ". OffsetY = ",  &
                      yoffset, ". One of these data must be wrong"
                call warning("spGatti11")
             else
                numb = numb + 1
             endif
          endif

! Keep only good events including those outside the field of view

          if (iand(qualev(jrow),testqual) == 0) then
             specgat(decal) = specgat(decal) + 1 
             spece1p(e1p)   = spece1p(e1p)   + 1
             spece1(ener1(jrow)) = spece1(ener1(jrow)) + 1
             maxdecal = max(maxdecal,decal)
          endif

       enddo
       count = count + nvalid(irow)
    enddo 

    write(errstr,*) numb, ' events with truncated E1'
    call message(AppMsg,Sparse)

    call release(e1_col)
    if (timeMode) then
       deallocate(ener1)
       call release(patt_col)
    endif

    if (outgatti) then
       spece1g = 0.
       do chan = 0, maxchan
          gat2 = min(chan,maxdecal)
          ch1  = chan - gat2
          spece1g(chan) = dot_product(specgat(gat2:0:-1),spece1(ch1:chan))
       enddo
       spece1g = spece1g / sum(spece1)
       call release(gat_set)
    else
       deallocate(spece1p)
    endif

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, spgatti_on)
    call setEventAttribute(ev_tab,EVATT_REJECTED_BY_GATTI)

  end subroutine spGatti

! Transpose the pattern library
! Set the 2s and the central 1 to 0

  subroutine preparePatterns(patterns,patabove)

! patterns : Pattern geometry given by the CAL
! patabove : Pattern geometry required by projectEventsEnergy

    integer(kind=int8), dimension(:,:,:), pointer  :: patterns
    integer(kind=int16),dimension(-1:1,-1:1,0:edu_npat-1),intent(out)::patabove

    integer :: npattern, jr

    patabove = 0

    do npattern = 0,size(patterns,1)-1
       do jr = -1,1
          patabove(:,jr,npattern) = patterns(npattern,jr+2,1:3)
       enddo
    enddo
    where (patabove > 1)
       patabove = 0
    endwhere
    patabove(0,0,:) = 0

  end subroutine preparePatterns

! Look for non-isolated (cosmic-ray) patterns
! In TIMING mode patterns are 1-D only, and include 1s in the borders

  subroutine cosmicrayPattern(patterns, cosmic)

! patterns  : Pattern library read from the CAL
! cosmic    : true for non-isolated patterns

    integer(kind=int8), dimension(:,:,:), pointer  :: patterns
    logical, dimension(:), pointer :: cosmic

    integer :: numpat, ie, ix, iy
    logical :: isolated

    numpat = size(patterns,1)
    allocate(cosmic(0:numpat-1))

    do ie = 0, numpat-1
       isolated = .true.

! Test whether they are isolated (no 0 around)

       if (imageMode .or. rimagMode) then
          do ix = 1,3
             do iy = 1,3
                if (patterns(ie,iy,ix) == 1) then
                   isolated = isolated .and.  &
                              minval(patterns(ie,iy-1:iy+1,ix-1:ix+1)) > 0
                endif
             enddo
          enddo
       else
          isolated = (patterns(ie,2,0) /= 1) .and. (patterns(ie,2,4) /= 1)
       endif

       cosmic(ie) = .not. isolated
    enddo

  end subroutine cosmicrayPattern

! Check validity of event coordinates and pattern

  subroutine checkValid(jrow, evtbuf, frameoff, x1, x2, y1, y2, p1, p2,  &
                        withwarn, orderbuf, nvalid, nevents3, noutside)

! jrow     : index for orderbuf
! evtbuf   : array of event characteristics
! frameoff : offset between frame number and index of nvalid
! x1       : minimum acceptable value for rawx
! x2       : maximum acceptable value for rawx
! y1       : minimum acceptable value for rawy
! y2       : maximum acceptable value for rawy
! p1       : minimum acceptable value for pattern
! p2       : maximum acceptable value for pattern
! withwarn : if true, send a warning when condition not fulfilled
! orderbuf : list of considered events
! nvalid   : number of events per frame
! nevents3 : current event number
! noutside : number of rejected events

    integer, intent(in)    :: jrow, frameoff, x1, x2, y1, y2, p1, p2
    logical, intent(in)    :: withwarn
    integer, intent(inout) :: nevents3
    integer, intent(inout), optional :: noutside
    type(sas_evt),       dimension(:), intent(in)    :: evtbuf
    integer(kind=int32), dimension(:), intent(inout) :: orderbuf
    integer(kind=int32), dimension(:), pointer       :: nvalid

    integer :: order, x, y, p, irow
    logical :: ok1, ok2, ok3

    order = orderbuf(jrow)
    x = evtbuf(order)%rawx
    y = evtbuf(order)%rawy
    p = evtbuf(order)%pattern
    ok1 = (x >= x1 .and. x <= x2)
    ok2 = (y >= y1 .and. y <= y2)
    ok3 = (p >= p1 .and. p <= p2)
    if (ok1 .and. ok2 .and. ok3)  then
       nevents3 = nevents3 + 1
       orderbuf(nevents3) = order
    else
       irow = evtbuf(order)%frame - frameoff
       nvalid(irow) = nvalid(irow) - 1
       if (present(noutside))  noutside = noutside + 1
       if (withwarn) then
          if (.not. ok1) then
             write(errstr,*) "Invalid event RAWX coordinate", x,  &
                             " at frame", evtbuf(order)%frame
          else if (.not. ok2) then
             write(errstr,*) "Invalid event RAWY coordinate", y,  &
                             " at frame", evtbuf(order)%frame
          else
             write(errstr,*) "Invalid event PATTERN value", p,  &
                             " at frame", evtbuf(order)%frame
          endif
          call warning("checkValid14")
       endif
    endif

  end subroutine checkValid

! Look for diagonal (two pixels touching by a corner) patterns

  subroutine diagonalPattern(patterns, diagonal)

! patterns  : Pattern library read from the CAL
! diagonal  : true for diagonal patterns

    integer(kind=int8), dimension(:,:,:), pointer  :: patterns
    logical, dimension(:), pointer :: diagonal

    integer, dimension(:), pointer :: diagpat
    integer :: numpat, numdiag, ie, ix, iy, iz, numdiagpat
    logical :: isolated

    numpat = size(patterns,1)
    allocate(diagonal(0:numpat-1), diagpat(numpat))
    numdiagpat = 0
    diagonal = .false.

    do ie = 0, numpat-1

! Test whether the pixels around are above threshold

       isolated = .true.
       do ix = 1,3,2
          isolated = isolated .and. patterns(ie,2,ix) == 2
          isolated = isolated .and. patterns(ie,ix,2) == 2
       enddo

! Test whether exactly one diagonal pixel is above threshold
! and that it is isolated (not corners)

       numdiag = 0
       do ix = 1,3,2
          do iy = 1,3,2
             if (patterns(ie,iy,ix) == 1) then
                do iz = -1,1,2
                   isolated = isolated .and. patterns(ie,iy,ix+iz) == 2
                   isolated = isolated .and. patterns(ie,iy+iz,ix) == 2
                enddo
                if (isolated)  numdiag = numdiag + 1
             endif
          enddo
       enddo

       diagonal(ie) = isolated .and. numdiag == 1
       if (diagonal(ie)) then
          numdiagpat = numdiagpat + 1
          diagpat(numdiagpat) = ie
       endif
    enddo

    write(errstr,*) "Diagonal patterns:", diagpat(1:numdiagpat)
    call message(AppMsg,Verbose)
    deallocate(diagpat)

  end subroutine diagonalPattern

! Conversion to detector coordinates

  subroutine getDetCoords(xin, yin, xout, yout, conv, focalLength, printon)

! xin,yin   : RAWX/RAWY coordinates
! xout,yout : DETX/DETY coordinates
! conv      : conversion factor from radians to detector coords
! focalLength: Focal length of the telescope
! printon   : sets verbosity on

    integer(kind=int16), dimension(:), pointer :: xin, yin, xout, yout
    real(kind=double), intent(in)              :: conv, focalLength
    logical,           intent(in)              :: printon

    real(double), dimension(:), pointer :: xmilli, ymilli, zmilli

    call CAL_rawXY2mm(xin,yin,xmilli,ymilli,zmilli,detSysType)

    if (printon) then
       write(errstr,*) "posx interval : ", minval(xin), maxval(xin)
       call message(AppMsg,Verbose)
       write(errstr,*) "posy interval : ", minval(yin), maxval(yin)
       call message(AppMsg,Verbose)
       write(errstr,*) "xmilli interval : ", minval(xmilli), maxval(xmilli)
       call message(AppMsg,Verbose)
       write(errstr,*) "ymilli interval : ", minval(ymilli), maxval(ymilli)
       call message(AppMsg,Verbose)
       write(errstr,*) "zmilli interval : ", minval(zmilli), maxval(zmilli)
       call message(AppMsg,Verbose)
    endif

! nint does not work with kind=int16 on NAGware
! Z axis of CAMCOORD2 points toward the mirror

    zmilli = focalLength - zmilli
    xmilli = det_offset + xmilli*conv/zmilli
    ymilli = det_offset + ymilli*conv/zmilli
    xout = int(xmilli+sign(0.5d0,xmilli), kind=int16)
    yout = int(ymilli+sign(0.5d0,ymilli), kind=int16)
    call CAL_releaseMemory(xmilli, ymilli, zmilli)

  end subroutine getDetCoords

! Get extreme values of detector coordinates

  subroutine getDetBorders(key1,key2,conv,focalLength,xmin,xmax,ymin,ymax)

! key1,key2 : Keyword names (TLMIN/TLMAX)
! conv      : conversion factor from mm to detector coords
! focalLength: Focal length of the telescope
! xmin,xmax : Min/max DETX value
! ymin,ymax : Min/max DETY value

    character(len=*),    intent(in)  :: key1, key2
    real(kind=double),   intent(in)  :: conv, focalLength
    integer(kind=int16), intent(out) :: xmin, xmax, ymin, ymax

    integer(kind=int16), dimension(:), pointer :: xcorn, ycorn, xout, yout

    if (hasAttribute(rawx_col,key1)) then
       xmin = int16Attribute(rawx_col,key1)
       xmax = int16Attribute(rawx_col,key2)
    else
       xmin = 1
       xmax = 600
    endif
    if (timeMode .or. ctimeMode) then
       ymin = EMOS_MAX_Y / 2
       ymax = EMOS_MAX_Y / 2
    else if (hasAttribute(rawy_col,key1)) then
       ymin = int16Attribute(rawy_col,key1)
       ymax = int16Attribute(rawy_col,key2)
    else
       ymin = 1
       ymax = 600
    endif

! Add 1 pixel margin, build the corners 
! and deduce extreme detector coordinates

    allocate(xcorn(4),ycorn(4),xout(4),yout(4))
    xcorn = (/xmin-1,xmax+1,xmax+1,xmin-1/)
    ycorn = (/ymin-1,ymin-1,ymax+1,ymax+1/)
    call CAL_setState(randomize=.false.)
    call getDetCoords(xcorn, ycorn, xout, yout, conv, focalLength, .false.)
    xmin = minval(xout)
    xmax = maxval(xout)
    ymin = minval(yout)
    ymax = maxval(yout)
    deallocate(xcorn,ycorn,xout,yout)

  end subroutine getDetBorders

! Find the interval for Si fluorescence

  subroutine initSifluor(patterns, diagonal, emini, emaxi,offpattern)

! patterns   : Pattern library read from the CAL
! diagonal   : true for diagonal patterns
! emini      : Lower bound of Si line interval (ADU)
! emaxi      : Upper bound of Si line interval (ADU)
! offpattern : Difference in pattern number between diagonals and bipixels

    integer(kind=int8), dimension(:,:,:), pointer  :: patterns
    logical, dimension(:), pointer :: diagonal
    integer, intent(out) :: emini, emaxi, offpattern

    integer(kind=int8), dimension(:,:,:), pointer :: patabove
    integer(kind=int8), dimension(:),     pointer :: pixelInE2
    integer(kind=int16), dimension(:),    pointer :: channels
    real(kind=double),  dimension(:),     pointer :: emin, emax, response
    real(kind=double)   :: respmax
    integer(kind=int16) :: c0
    integer :: bipat1, bipat2, diagpat1, diagpat2, patcur, numpat
    integer :: phamax, phacur
    logical :: firstbi, firstdiag

! Average energy for Si K line

    real(kind=double), parameter :: enersik = 1740.

! Get the number of pixels making E2

    numpat = size(patterns,1)
    allocate(patabove(0:numpat-1,3,3),pixelInE2(0:numpat-1))
    patabove = patterns(:,1:3,1:3)
    patabove(:,2,2) = 0
    where (patabove > 1)
       patabove = 0
    endwhere
    pixelInE2 = sum(sum(patabove,3),2)

! Deduce the difference between bipixels and diagonals

    bipat1 = 0
    bipat2 = 0
    diagpat1 = 0
    diagpat2 = 0
    firstbi = .true.
    firstdiag = .true.
    do patcur = 0,31
       if (diagonal(patcur)) then
          if (firstdiag)  diagpat1 = patcur
          diagpat2 = patcur
          firstdiag = .false.
       else if (pixelInE2(patcur) == 1) then
          if (firstbi)  bipat1 = patcur
          bipat2 = patcur
          firstbi = .false.
       endif
    enddo
    if ((diagpat2-diagpat1)/=3 .or. (bipat2-bipat1)/=3) then
       write(errstr,*) "Bipixels or diagonals do not have consecutive ",  &
          "pattern numbers. Si line will remain diagonal."
       call warning("initSifluor11")
       offpattern = 0
    else
       offpattern = diagpat1 - bipat1
    endif

    deallocate(patabove,pixelInE2)

! Get instrument response at energy of Si K
! Interval is C0 +/- HWHM in PI.

    call CAL_getEbounds(channels, emin, emax)
    call CAL_getRedistribution(enersik, emin, emax, response, GRADE_SINGLE, c0)

! Look for maximum, starting at c0

    phamax = c0
    respmax = response(phamax)
    do phacur = phamax+1, size(response)-1
       if (response(phacur) > respmax) then
          respmax = response(phacur)
       else
          exit
       endif
    enddo
    if (phacur > phamax+1) then   ! The maximum was above c0
       phamax = phacur - 1
    else                         ! The maximum may be below c0
       do phacur = phamax-1, 0, -1
          if (response(phacur) > respmax) then
             respmax = response(phacur)
          else
             exit
          endif
       enddo
       if (phacur < phamax-1)  phamax = phacur + 1
    endif

! Look for interval at half max

    do emaxi = phamax+1, size(response)-1
       if (response(emaxi) < 0.5d0*respmax) exit
    enddo

    do emini = phamax-1, 0, -1
       if (response(emini) < 0.5d0*respmax) exit
    enddo

    call CAL_releaseMemory(channels)
    call CAL_releaseMemory(emin, emax, response)

    write(errstr,*) "The ADU interval for Si line is", emini, emaxi
    call message(AppMsg,Verbose)

  end subroutine initSifluor

! Copy event in output of EDUSOFT

  subroutine copyEvent(inrow, outrow, moveX, moveY, eduevt, evtbuf)

    integer, intent(in) :: inrow, outrow, moveX, moveY
    type(edu_evt), intent(in) :: eduevt
    type(sas_evt), dimension(:), intent(inout) :: evtbuf

    evtbuf(outrow) = evtbuf(inrow)
    evtbuf(outrow)%rawx    = eduevt%x + moveX
    evtbuf(outrow)%rawy    = eduevt%y + moveY
    evtbuf(outrow)%pattern = eduevt%pattern
    evtbuf(outrow)%e1      = eduevt%e1
    evtbuf(outrow)%e2      = eduevt%e2

  end subroutine copyEvent

! Get threshold and edumode for edusoft

  subroutine initEdusoft(ev_tab, threshold, edumode)

    type(TableT),        intent(in)  :: ev_tab
    integer(kind=int16), intent(out) :: threshold
    integer,             intent(out) :: edumode

    threshold = int16Attribute(ev_tab,"EDUTHR")
    if (threshold < 1) then
       write(errstr,*) "EDUTHR =", threshold, " should be positive."
       call warning("initEdusoft10")
    else
       write(errstr,*) "EDUTHR =", threshold
       call message(AppMsg,Verbose)
    endif
    edumode = int32Attribute(ev_tab,"EDUMODE")
    if (edumode < 1 .or. edumode > 3) then
       write(errstr,*) "EDUMODE =", edumode, " should be 1, 2 or 3."
       call error("initEdusoft02")
    else if (imageMode .and. edumode /= 3) then
       write(errstr,*) "EDUMODE =", edumode, " should be 3 in IMAGING mode."
       call error("initEdusoft03")
    else if ((timeMode.or.ctimeMode) .and. edumode /= 1) then
       write(errstr,*) "EDUMODE =", edumode, " should be 1 in TIMING mode."
       call error("initEdusoft04")
    else if (rimagMode .and. edumode /= 2) then
       write(errstr,*) "EDUMODE =", edumode, " is not supported ",  &
                       "in REDUCED IMAGING mode. Should be 2 (Threshold)."
       call error("initEdusoft05")
    else
       write(errstr,*) "EDUMODE =", edumode
       call message(AppMsg,Verbose)
       if (rimagMode)  edumode = 3     ! EDU needs to be applied
    endif

  end subroutine initEdusoft

! Check the proportion of rejected events

  subroutine checkRejected(ev_tab, neventsin)

! ev_tab    : Handle to the event table
! neventsin : Original number of events

    type(TableT), intent(in) :: ev_tab
    integer, intent(in)      :: neventsin

    real    :: frac
    integer :: neventsout, jrow

    if (neventsin < 1)  return
    neventsout = 0
    do jrow = 0, numberOfRows(ev_tab)-1
       if (iand(qualev(jrow),EVATT_EMOS_REJECTION_MASK) == 0)  &
          neventsout = neventsout + 1
    enddo

    frac = real(neventsout) / real(neventsin)
    write(errstr,*) "Fraction of good events : ", frac
    call message(AppMsg,Sparse)
    if (frac < fracmin) then
       write(errstr,*) "the fraction of good events ", frac,  &
                       " is less than ", fracmin
       call warning("checkRejected10")
    endif

  end subroutine checkRejected

! Read the BADPIX extension

  subroutine readBadpix(bad1_tab, badx, bady, badtype, badext, badflag)

! bad1_tab : Handle to the input badpix table
! others   : Pointers to the data

    type(TableT), intent(in)   :: bad1_tab
    integer(kind=int16), dimension(:), pointer :: badx, bady, badtype,  &
                                                  badext, badflag

    type(ColumnT) :: badx1_col, bady1_col, badt1_col, bade1_col, badf1_col
    integer :: jbad, ybad

! Read the columns

       badx1_col = column(bad1_tab, "RAWX", Read)
       bady1_col = column(bad1_tab, "RAWY", Read)
       badt1_col = column(bad1_tab, "TYPE", Read)
       bade1_col = column(bad1_tab, "YEXTENT", Read)
       badf1_col = column(bad1_tab, "BADFLAG", Read)
       badx    => int16Data(badx1_col)
       bady    => int16Data(bady1_col)
       badtype => int16Data(badt1_col)
       badext  => int16Data(bade1_col)
       badflag => int16Data(badf1_col)

! Check that no bad pixel has invalid coordinates (RAWX=0 is valid)
! If they do set to acceptable coordinates outside the CCD
! Remove dead and intact pixels (not useful to CUT_BAD)

       do jbad = 0,size(badx)-1
          ybad = bady(jbad)
          if (min(badx(jbad)+1,ybad) < 1 .or.  &
              min(x_size_window-underx-overx-int(badx(jbad)),  &
                y_size_window-undery-overy-int(ybad+badext(jbad))+1) < 0) then
              write(errstr,*) "Invalid bad pixel coordinates", badx(jbad), &
                   ",", ybad, ",", ybad+badext(jbad)-1, " at row", jbad+1
              call warning("readBadpix15")
              badx(jbad) = 0
              bady(jbad) = 0
              badext(jbad) = 1
          endif
          if ((badtype(jbad) == INTACT) .or. (badtype(jbad) == DEAD)) then
              badx(jbad) = 0
              bady(jbad) = 0
              badext(jbad) = 0
          endif
       enddo

  end subroutine readBadpix

! Declare all columns of the event list
! Note that some columns are defined as global variables

  subroutine declare_all_columns(ev_tab,e2_col,e3_col,e4_col,peri_col)

! ev_tab   : Handle to the event table
! xx_col   : Handle to the columns

    type(TableT), intent(in)   :: ev_tab
    type(ColumnT), intent(out) :: e2_col, e3_col, e4_col, peri_col
    integer :: eduthr, emdhlow, emdhupp

    frev_col = addColumn(ev_tab, fram, Integer32, "COUNTER",  &
                         comment="frame number in the auxiliary file")
    flev_col = addColumn(ev_tab, EVATT_COLUMN_NAME, Integer32, "",  &
                         comment="quality flag")
    rawx_col = addColumn(ev_tab, rawx, Integer16, "PIXELS",  &
                         comment="X in node coordinates")
    rawy_col = addColumn(ev_tab, rawy, Integer16, "PIXELS",  &
                         comment="Y in node coordinates")
    patt_col = addColumn(ev_tab, patt, Integer8,  "",  &
                      comment="pattern or grade defining the event's geometry")
    call setAttribute(patt_col, "TLMIN",  0, comment="Minimum for "//patt)
    call setAttribute(patt_col, "TLMAX", 31, comment="Maximum for "//patt)
    peri_col = addColumn(ev_tab, peri, Integer8,  "PIXELS",  &
                         comment="number of pixels > EDUTHR in outer corona")
    call setAttribute(peri_col, "TLMIN",  0, comment="Minimum for "//peri)
    call setAttribute(peri_col, "TLMAX",  7, comment="Maximum for "//peri)

! Define energy bounds
! EMDHUPP should never be larger than the data format allows

    if (hasAttribute(ev_tab,"EDUTHR")) then
       eduthr = int32Attribute(ev_tab,"EDUTHR")
    else
       eduthr = 0
    endif
    if (hasAttribute(ev_tab,"EMDHLOW"))  &
       emdhlow = int32Attribute(ev_tab,"EMDHLOW")
    emdhlow = max(eduthr,emdhlow)
    emdhupp = maxchan
    if (hasAttribute(ev_tab,"EMDHUPP"))  &
       emdhupp = min(32767,int32Attribute(ev_tab,"EMDHUPP"))

    if (imageMode .or. rimagMode) then
       e1_col = addColumn(ev_tab, ene1, Integer16, ener_unit,  &
                          comment="charge in the central pixel")
       e2_col = addColumn(ev_tab, ene2, Integer16, ener_unit,  &
                          comment="total charge in inner pixels >= EDUTHR")
       call setAttribute(e1_col, "TLMIN",      0,  &
                         comment="Legal minimum for "//ene1)
       call setAttribute(e1_col, "TDMIN", eduthr,  &
                         comment="Actual minimum for "//ene1)
       call setAttribute(e1_col, "TLMAX",maxchan,  &
                         comment="Legal maximum for "//ene1)
       call setAttribute(e1_col, "TDMAX",min(maxchan,emdhupp),  &
                         comment="Actual maximum for "//ene1)
       call setAttribute(e2_col, "TLMIN",     0,  &
                         comment="Legal minimum for "//ene2)
       call setAttribute(e2_col, "TDMIN", eduthr,  &
                         comment="Actual minimum for "//ene2)
       call setAttribute(e2_col, "TLMAX",  32767,  &
                         comment="Legal maximum for "//ene2)
       call setAttribute(e2_col, "TDMAX",emdhupp-eduthr,  &
                         comment="Actual maximum for "//ene2)
       if (imageMode) then
          e3_col = addColumn(ev_tab, ene3, Integer16, ener_unit,  &
                             comment="total charge in inner pixels < EDUTHR")
          e4_col = addColumn(ev_tab, ene4, Integer16, ener_unit,  &
                             comment="total charge in outer pixels < EDUTHR")
       endif
    else if (timeMode .or. ctimeMode) then

! Directly rename ENGYE1E2 to PHA
! Define Nulls in Compressed Timing

       e1_col = addColumn(ev_tab, pha, Real32, ener_unit,  &
                          comment="total charge of the event (uncorrected)")
       call setAttribute(e1_col, "TLMIN",      0,  &
                         comment="Legal minimum for "//pha)
       call setAttribute(e1_col, "TDMIN",emdhlow,  &
                         comment="Actual minimum for "//pha)
       call setAttribute(e1_col, "TLMAX",maxchan,  &
                         comment="Legal maximum for "//pha)
       call setAttribute(e1_col, "TDMAX",emdhupp,  &
                         comment="Actual maximum for "//pha)
       if (ctimeMode) then
          call setNullValue(rawx_col,-1)
          call setNullValue(patt_col,255)
          call setNullValue(peri_col,255)
       endif
    endif

  end subroutine declare_all_columns

! Enter in memory all columns of the event list for writing
! Note that some columns are defined as global variables

  subroutine enter_all_columns(ev_tab,e2_col,e3_col,e4_col,peri_col,  &
                               periout,e2out,e3out,e4out,phaout)

! ev_tab   : Handle to the event table
! xx_col   : Handle to the columns
! xxout    : Pointers to the data

    type(TableT), intent(in)   :: ev_tab
    type(ColumnT), intent(out) :: e2_col, e3_col, e4_col, peri_col
    integer(kind=int8),  dimension(:), pointer :: periout
    integer(kind=int16), dimension(:), pointer :: e2out, e3out, e4out
    real(kind=single),   dimension(:), pointer :: phaout

    frev_col = column(ev_tab, fram, Modify)
    flev_col = column(ev_tab, EVATT_COLUMN_NAME, Modify)
    rawx_col = column(ev_tab, rawx, Modify)
    rawy_col = column(ev_tab, rawy, Modify)
    patt_col = column(ev_tab, patt, Modify)
    peri_col = column(ev_tab, peri, Modify)
    frameev =>  int32Data(frev_col)
       posx =>  int16Data(rawx_col)
       posy =>  int16Data(rawy_col)
     qualev =>  int32Data(flev_col)
    pattern =>   int8Data(patt_col)
    periout =>   int8Data(peri_col)

    if (imageMode .or. rimagMode) then
       e1_col = column(ev_tab, ene1, Modify)
       e2_col = column(ev_tab, ene2, Modify)
       ener1 => int16Data(e1_col)
       e2out => int16Data(e2_col)
       if (imageMode) then
          e3_col = column(ev_tab, ene3, Modify)
          e4_col = column(ev_tab, ene4, Modify)
          e3out => int16Data(e3_col)
          e4out => int16Data(e4_col)
       endif
    else if (timeMode .or. ctimeMode) then
       e1_col = column(ev_tab, pha, Modify)
       phaout => real32Data(e1_col)
    endif

  end subroutine enter_all_columns

! Release all columns of the event list

  subroutine release_all_columns(e2_col,e3_col,e4_col,peri_col)

! xx_col   : Handle to the columns

    type(ColumnT), intent(in)  :: e2_col, e3_col, e4_col, peri_col

    call release(frev_col)
    call release(flev_col)
    call release(rawx_col)
    call release(rawy_col)
    call release(patt_col)
    call release(peri_col)
    call release(  e1_col)
    if (imageMode .or. rimagMode) then
       call release(e2_col)
       if (imageMode) then
          call release(e3_col)
          call release(e4_col)
       endif
    endif

  end subroutine release_all_columns

! Add the new columns for future use and set them to 0

  subroutine add_new_columns(ev_tab)

! ev_tab   : Handle to the event table

    type(TableT), intent(in)   :: ev_tab

    type(ColumnT) :: detx_col,dety_col,x_col,y_col,pha_col,pi_col,tiev_col
    integer(kind=int16), dimension(:), pointer :: xdet,ydet
    integer(kind=int32), dimension(:), pointer :: x,y
    real(kind=single),   dimension(:), pointer :: phaout,piout
    real(kind=double),   dimension(:), pointer :: timeout

    tiev_col = addColumn(ev_tab, time, Real64,    time_unit,  &
                         comment="time of the center of the frame")
    timeout => real64Data(tiev_col)
    timeout = 0.d0
    call release(tiev_col)
      pi_col = addColumn(ev_tab, pi,   Real32,    ener_unit,  &
                       comment="measured energy in eV (gain corrected)")
      piout => real32Data(  pi_col)
      piout = 0.
    call release(  pi_col)

! DETX/DETY, X/Y are now generated in Timing mode also

    detx_col = addColumn(ev_tab, detx, Integer16, coord_unit,  &
                         comment="X in linearised camera coordinates")
    dety_col = addColumn(ev_tab, dety, Integer16, coord_unit,  &
                         comment="Y in linearised camera coordinates")
       x_col = addColumn(ev_tab, skyx, Integer32, coord_unit,  &
                         comment="X in projected sky coordinates")
       y_col = addColumn(ev_tab, skyy, Integer32, coord_unit,  &
                         comment="Y in projected sky coordinates")
      xdet =>  int16Data(detx_col)
      ydet =>  int16Data(dety_col)
         x =>  int32Data(   x_col)
         y =>  int32Data(   y_col)
      xdet = 0
      ydet = 0
         x = 0
         y = 0
    call release(detx_col)
    call release(dety_col)
    call release(   x_col)
    call release(   y_col)

! In Timing mode PHA is directly copied from ENGYE1E2

    if (imageMode .or. rimagMode) then
       pha_col = addColumn(ev_tab, pha,  Real32,    ener_unit,  &
                           comment="total charge of the event (uncorrected)")
       phaout => real32Data( pha_col)
       phaout = 0.
       call release( pha_col)
    endif

  end subroutine add_new_columns

! Read part of the event list into the buffer

  subroutine getEvents(nevents, event0, ev_tab, evtbuf)

! nevents : number of rows
! event0  : first row to read
! ev_tab  : handle to the input table for the DAL
! evtbuf  : event structure for output

    integer,       intent(in)  :: nevents, event0
    type(TableT),  intent(in)  :: ev_tab
    type(sas_evt), dimension(:), intent(out) :: evtbuf

    real(kind=double), dimension(:), allocatable :: timebuf

    allocate(timebuf(size(evtbuf)))

    timebuf = 0.d0
    call copyColumnPart(ev_tab, fram, nevents, event0, timebuf)
    evtbuf%frame = nint(timebuf)
    call copyColumnPart(ev_tab, rawx, nevents, event0, timebuf)
    evtbuf%rawx = int(timebuf+sign(0.5D0,timebuf),int16)
    call copyColumnPart(ev_tab, rawy, nevents, event0, timebuf)
    evtbuf%rawy = int(timebuf+sign(0.5D0,timebuf),int16)
    call copyColumnPart(ev_tab, patt, nevents, event0, timebuf)
    evtbuf%pattern = int(timebuf+sign(0.5D0,timebuf),int8)
    call copyColumnPart(ev_tab, peri, nevents, event0, timebuf)
    evtbuf%peripix = int(timebuf+sign(0.5D0,timebuf),int8)
    if (imageMode) then
       call copyColumnPart(ev_tab, ene1, nevents, event0, timebuf)
       evtbuf%e1 = int(timebuf+sign(0.5D0,timebuf),int16)
       call copyColumnPart(ev_tab, ene2, nevents, event0, timebuf)
       evtbuf%e2 = int(timebuf+sign(0.5D0,timebuf),int16)
       call copyColumnPart(ev_tab, ene3, nevents, event0, timebuf)
       evtbuf%e3 = int(timebuf+sign(0.5D0,timebuf),int16)
       call copyColumnPart(ev_tab, ene4, nevents, event0, timebuf)
       evtbuf%e4 = int(timebuf+sign(0.5D0,timebuf),int16)
    else if (timeMode .or. ctimeMode .or. rimagMode) then
       call copyColumnPart(ev_tab, en12, nevents, event0, timebuf)
       evtbuf%e1 = int(timebuf+sign(0.5D0,timebuf),int16)
       evtbuf%e2 = 0
    endif
    evtbuf%flag = 0

    deallocate(timebuf)

  end subroutine getEvents

! Copy part of a column into array with possible type change

  subroutine copyColumnPart(ev_tab, colname, nrows, ncur, buffer)

! ev_tab  : handle to the input table for the DAL
! colname : column name
! nrows   : number of rows in input column
! ncur    : which row to start in the input column
! buffer  : output array (double)

    type(TableT),  intent(in)  :: ev_tab
    integer, intent(in) :: nrows, ncur
    character(len=*), intent(in) :: colname
    real(kind=double), dimension(:), intent(out) :: buffer

    type(ColumnT) :: in_col
    integer :: coltype

! Possible types of columns

    real(kind=double),   dimension(:), pointer   :: dbledata
    real(kind=single),   dimension(:), pointer   :: sngldata
    integer(kind=int32), dimension(:), pointer   :: longdata
    integer(kind=int16), dimension(:), pointer   :: intgdata
    integer(kind=int8),  dimension(:), pointer   :: bytedata

! Read input column

    in_col  = column(ev_tab, colname, Read)
    coltype = columnDataType(in_col)
    select case(coltype)
    case (Real64)
       dbledata => real64Data(in_col)
       buffer(1:nrows) = dbledata(ncur:ncur+nrows-1)
    case (Real32)
       sngldata => real32Data(in_col)
       buffer(1:nrows) = sngldata(ncur:ncur+nrows-1)
    case (Integer32)
       longdata => int32Data(in_col)
       buffer(1:nrows) = longdata(ncur:ncur+nrows-1)
    case (Integer16)
       intgdata => int16Data(in_col)
       buffer(1:nrows) = intgdata(ncur:ncur+nrows-1)
    case (Integer8)
       bytedata => int8Data(in_col)
       buffer(1:nrows) = bytedata(ncur:ncur+nrows-1)
    case default
       write(errstr,*) "input column ", trim(qualifiedName(in_col)),  &
                       " has an unknown type ", coltype
       call error("copyColumnPart01")
    end select
    call release(in_col)

  end subroutine copyColumnPart

! Open the output event file

  subroutine openEventsOut(event_out, ev1_tab, ev_tab)

! event_out : Name of the output event file
! ev1_tab   : Handle to the input event table
! ev_tab    : Handle to the output event table

    character(len=*), intent(in) :: event_out
    type(TableT), intent(inout)  :: ev1_tab
    type(TableT), intent(out)    :: ev_tab

    character(len=1024) :: event_in
    real                :: deltat
    type(DataSetT)      :: ev_set, ev1_set
    type(TableT)        :: bad1_tab

    call message(AppMsg,Verbose,"Opening output event file ...")

    ev1_set = parent(ev1_tab)
    if (overevent) then
       ev_set = ev1_set
       ev_tab = ev1_tab
    else if (.not. fromodf) then
       event_in = name(ev1_set)
       call release(ev1_set)

!  Clone the input event set and define again ev_tab with the new dataSet
      
       write(errstr,*) "Input file ", trim(event_in), " copied to ", &
                       trim(event_out)
       call message(AppMsg,Sparse)
       ev_set = clone(event_in, event_out, MODIFY)
       ev_tab = table(ev_set, event_ext)
       ev1_tab = ev_tab
    else
       ev_set = dataSet(event_out, Create)
       ev_tab = addTable(ev_set, event_ext, 0,  &
                         "CCD/node specific event list")
       call setAttributes(attributable(ev_tab), attributable(ev1_tab))
       if ((imageMode.or.rimagmode) .and. hasAttribute(ev_tab, "FRMTIME")) then
          deltat = real32Attribute(ev_tab, "FRMTIME")
          deltat = deltat * 0.1      ! FRMTIME is in units of 100 ms
          call setAttribute(ev_tab, "FRMTIME", 1.e3*deltat, "ms",  &
                            comment="Frame integration time (command)")
       endif
       call addFilename(block(ev_tab),name(ev1_set),"Input event file :")

! Propagate the BADPIX extension if any

       if (hasBlock(ev1_set,badpix_ext)) then
          bad1_tab = table(ev1_set,badpix_ext)
          call copyBlock(ev_set, block(bad1_tab), badpix_ext)
          call release(bad1_tab)
          call message(AppMsg,Verbose,"Bad pixels extension propagated ...")
       endif
    endif

  end subroutine openEventsOut

! Add the offsets to the output event file

  subroutine addOffsets(ev_tab, off_tab)

! ev_tab    : Handle to the output event table
! off_tab   : Handle to the input offset/variance table

    type(TableT), intent(in)     :: ev_tab
    type(TableT), intent(inout)  :: off_tab

    type(DataSetT)      :: ev_set
    type(ColumnT)       :: offx_col, offy_col

    ev_set = parent(ev_tab)
    if (fromodf) then

! Propagate the OFFSETS extension if any
! Read the offsets from CAL if no offsets file is provided

       if (onoffvar) then
          call copyBlock(ev_set, block(off_tab), offset_ext)
          call message(AppMsg,Verbose,"Offset/variance extension propagated.")
          call release(parent(off_tab))
          call patchOffsets(ev_tab)
       else
          call getCalOffsets(ev_tab)
       endif
    endif

! Case when emevents is restarted on its own output

    onoffvar = onoffvar .or. hasBlock(ev_set, offset_ext)

! Check that offsets are available when they are required

    if (onoffvar) then     ! Read the offsets
       write(errstr,*) "Reading the offset/variance extension ..."
       call message(AppMsg,Verbose)
       off_tab = table(ev_set, offset_ext)
       offx_col = column(off_tab, "COLOFFST", Read)
       offsetX => int16VectorData(offx_col)
       offy_col = column(off_tab, "ROWOFFST", Read)
       offsetY => int16VectorData(offy_col)
    else
       if (onspgatti)  then
          call warning("openEventsOut14",  &
             "SP_GATTI requires an offset/variance file. Not applied")
          onspgatti = .false.
       endif
       if (oncutbad .or. onbinodal .or. onputXY .or. e3reject .or.  &
           rimagMode) then
          call warning("openEventsOut15",  &
             "Offsets will be fixed to 100 for CUT_BAD, BINODAL and REJECT_E3")
          allocate(offsetX(0:1023,0:0),offsetY(0:1023,0:0))
          offsetX = 50
          offsetY = 50
       endif
    endif

  end subroutine addOffsets

! Open the input event file

  subroutine openEventsIn(event_in,offvar,off_tab,ev1_tab,neventsin)

! event_in  : Name of the input event file
! offvar    : Name of the offset/variance files list
! off_tab   : Handle to the input offset/variance table
! ev1_tab   : Handle to the input event table
! neventsin : Number of events in input

    character(len=*), intent(in)                 :: event_in
    character(len=*), dimension(:), intent(in)   :: offvar
    type(TableT), intent(out)                    :: ev1_tab, off_tab
    integer, intent(out)                         :: neventsin

    type(DataSetT)      :: ev1_set

    call message(AppMsg,Verbose,"Opening input event file ...")

    if (overevent) then
       ev1_set = dataSet(event_in, Modify)
    else
       ev1_set = dataSet(event_in, Read)
    endif

    fromodf = .not. hasBlock(ev1_set, event_ext)

! The input file is straight from the ODF

    if (fromodf) then

       if (overevent) then
          write(errstr,*) "The event file ", trim(name(ev1_set)),  &
             " doesn't have the correct structure for overwriting. No ",  &
             trim(event_ext), " extension. The overwrite option can be ",  &
             "used only on the output of a previous run of emevents."
          call error("openEventsIn01")
       endif
       ev1_tab = table(ev1_set, 0)  ! Extension name depends on mode
       call checkExtension(ev1_tab)

       if (onoffvar) then
          call checkOffsets(ev1_tab, offvar, off_tab)
       endif

! The input file was already processed by emevents

    else
       ev1_tab = table(ev1_set, event_ext)
       if (onoffvar) then
          write(errstr,*) "It is not allowed to change the offsets ", &
                          "mid-way through emevents. The offset/variance ", &
                          "files will be ignored."
          call warning("openEventsIn14")
          onoffvar = .false.
       endif
    endif

    call getMode(ev1_tab, imageMode, timeMode, rimagMode, ctimeMode)

    neventsin = numberOfRows(ev1_tab)
    if (neventsin > 0) then
       write(errstr,*) neventsin, " events in input"
       call message(AppMsg,Sparse)
    else
       write(errstr,*) "the event file is empty."
       call warning("openEventsIn12")
    endif

  end subroutine openEventsIn

! Check extension name for ODF files

  subroutine checkExtension(ev_tab)

! ev_tab   : Handle to the input table

    type(TableT), intent(in) :: ev_tab

    character(len=16)   :: name_ext

    name_ext = name(ev_tab)
    if (name_ext(5:5) /= 'E') then
       write(errstr,*) "The extension name ", trim(qualifiedName(ev_tab)),  &
                       " tends to indicate that this is not an event ",  &
                       "list file. Should be something like MnIMEn."
       call error("checkExtension02")
    endif
    if (name_ext(1:1) /= 'M') then
       write(errstr,*) "The extension name ", trim(qualifiedName(ev_tab)),  &
                       " tends to indicate that this is not an EPIC MOS",  &
                       " file. Should be something like MnIMEn."
       call error("checkExtension03")
    endif

  end subroutine checkExtension

! Open the secondary input event file

  subroutine openEvents2In(event_in, ev1_tab, ev2_tab)

! event_in  : Name of the input event file
! ev1_tab   : Handle to the input event table
! ev2_tab   : Handle to the secondary event table

    character(len=*), intent(in) :: event_in
    type(TableT), intent(in)     :: ev1_tab
    type(TableT), intent(out)    :: ev2_tab
  
! Keywords to check for compatibility of both event files

    character(len=8), dimension(5), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","OBS_ID  ","EXP_ID  ","CCDID   "/)

    type(DataSetT)      :: ev2_set

    call message(AppMsg,Verbose,"Opening secondary input event file ...")

! The input file must be straight from the ODF

    ev2_set = dataSet(event_in, Read)
    ev2_tab = table(ev2_set, 0)  ! Extension name depends on mode
    call checkExtension(ev2_tab)

    if (.not. equalKeywords(attributable(ev2_tab),  &
                            attributable(ev1_tab), keycompat)) then
       call warning("openEvents2In13", "BINODAL will not be applied")
       onbinodal = .false.
       call release(ev2_set)
    endif

  end subroutine openEvents2In

! Read the input offset/variance file

  subroutine checkOffsets(ev1_tab, offvar, off_tab)

! ev1_tab   : Handle to the input event table
! offvar    : List of offset/variance files
! off_tab   : Handle to the input offset/variance table

    type(TableT), intent(in)                     :: ev1_tab
    character(len=*), dimension(:), intent(in)   :: offvar
    type(TableT), intent(out)                    :: off_tab

! Keywords to check for compatibility of offset/variance and event files
! Do not check obs. and exp. number (could use file from other obs.)
! FRMTIME is dealt with separately because in free run modes the PMS
! sets it to a wrong value in the offset/variance files
! WINDOWDY is dealt with separately because in Timing modes the PMS
! sets it to 1024 in the events files

    character(len=8), dimension(7), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","CCDID   ","CCDNODE ",&
           "WINDOWX0","WINDOWY0","WINDOWDX"/)
    character(len=8), dimension(2), parameter :: keycompat2 = &
         (/"FRMTIME ","WINDOWDY"/)
    type(DataSetT)      :: off_set
    type(ColumnT)       :: offx_col, offy_col
    type(TableT), dimension(:), allocatable :: goodtables
    real(double), dimension(:), allocatable :: starttimes
    integer,      dimension(:), allocatable :: better
    real(double)        :: start0, start1, startcur
    character(len=16)   :: name_ext
    logical             :: setok
    integer             :: nofva, ngood, nbetter, nbest

    nofva = size(offvar)
    allocate(goodtables(nofva),starttimes(nofva),better(nofva))
    starttimes = 0.d0
    start0 = getXmmDate(ev1_tab)
    ngood = 0
    nbetter = 0
    nbest = 0
    start1 = 0.d0

! Loop on files of the list 
! As soon as file found not ok, skip following tests and loop to the next file

    do nofva=1, size(offvar)
 
       setok = .true.

! Check existence of the file, if not cycle to next file

       if (.not. setExists(offvar(nofva))) then
          write(errstr,*) trim(offvar(nofva))," does not exist"
          call warning("checkOffsets10")
          setok = .false.
          cycle
       endif  
       off_set = dataSet(offvar(nofva), Read)
       write(errstr,*) "Check offset/variance file : ", trim(offvar(nofva))
       call message(AppMsg,Noisy)

       off_tab = table(off_set, 0)   ! Extension name is variable
       name_ext = name(off_tab)
       if (name_ext(1:1) /= 'M' .or. name_ext(3:5) /= 'OVE') then
          write(errstr,*) "Non expected extension ", trim(name(off_tab)),  &
             " in input offset/variance file ", trim(name(off_set)),  &
             ". Should be something like MnOVEn. File not used"
          call message(AppMsg,Noisy)
          setok = .false.
          call release(off_set)
          cycle
       endif

! Look for last offsets file before events file

       startcur = getXmmDate(off_tab)
       if (startcur > start1 .and. startcur < start0) then
          start1 = startcur
          nbest = 0
       endif

! Test the compatibility with the event file

       if (.not. equalKeywords(attributable(off_tab), attributable(ev1_tab),  &
                               keycompat, onwarn=.false.)) then
          write(errstr,*) "offset/variance file ", trim(name(off_set)), &
                          " is not compatible with the event file"
          call message(AppMsg,Noisy)
          setok = .false.
          call release(off_set)
          cycle
       endif

! Check offset sizes in column and row 

       offx_col = column(off_tab, "COLOFFST", Read)
       offsetX => int16VectorData(offx_col)
       offy_col = column(off_tab, "ROWOFFST", Read)
       offsetY => int16VectorData(offy_col)
       if (size(offsetX)<x_size_window.or.size(offsetY)<y_size_window) then
          write(errstr,*) "The offsets are incomplete. They have size",  &
              size(offsetX),",",size(offsetY)," <",x_size_window,",",  &
              y_size_window," in file ",trim(name(off_set)),". File not used."
          call message(AppMsg,Noisy)
          setok = .false.
          call release(off_set)
          cycle
       endif
       call release(offx_col)
       call release(offy_col)

! Build the list of compatible files and start times

       if ( setok ) then
          ngood = ngood + 1
          goodtables(ngood) = off_tab
          starttimes(ngood) = startcur
          if (abs(startcur-start1) < 1.d0)  nbest = ngood

! Check for perfect compatibility

          if (equalKeywords(attributable(off_tab), attributable(ev1_tab),  &
                            keycompat2, onwarn=.false.)) then
             nbetter = nbetter + 1
             better(nbetter) = ngood
          endif
       endif
    enddo

! Take the latest start time before current exposure if compatible
! Otherwise find the latest compatible start time before current exposure
! or the earliest compatible start time after current exposure
! Try first files with full compatibility then partial compatibility

    if ( ngood > 0 ) then
       if (nbest == 0) then
          if ( nbetter > 0 ) then
             call findBestTime(starttimes(better(1:nbetter)), start0, nbest)
             nbest = better(nbest)
          else
             call findBestTime(starttimes(1:ngood), start0, nbest)
          endif
       endif
       off_tab = goodtables(nbest)

! Release all files not selected

       do nofva = 1,ngood
          if (nofva /= nbest)  call release(parent(goodtables(nofva)))
       enddo

! Result of the tests 

       write(errstr,*) "offset/variance file ", trim(qualifiedName(off_tab)), &
                       " was selected "
       call message(AppMsg,Sparse)
    else
       write(errstr,*) "No offset/variance file was found correct"
       call warning("checkOffsets11")
       onoffvar = .false.
    endif

    deallocate(goodtables,starttimes)

  end subroutine checkOffsets

! Find the closest offset/variance file before events file

  subroutine findBestTime(starttimes, start0, nbest)

! starttimes : times of the offset/variance files
! start0     : time of the events file
! nbest      : index of best file

    real(double), dimension(:), intent(in)  :: starttimes
    real(double),               intent(in)  :: start0
    integer,                    intent(out) :: nbest

    integer :: ngood, nofva

    ngood = size(starttimes)
    if (ngood == 0)  call error("findBestTimes01",  &
                             "Wrong logic in checkOffsets. Contact developer")

    nbest = 0
    do nofva = 1,ngood
       if (starttimes(nofva) <= start0) then
          if (nbest == 0) then
             nbest = nofva
          else if (starttimes(nofva) > starttimes(nbest)) then
             nbest = nofva
          endif
       endif
    enddo
    if (nbest == 0) then
       write(errstr,*) "No offset/variance file was obtained earlier ",  &
                       "than the current exposure. Use first one obtained."
       call warning("findBestTimes12")
       nbest = 1
       do nofva = 2,ngood
          if (starttimes(nofva) < starttimes(nbest))  nbest = nofva
       enddo
    endif

  end subroutine findBestTime

! Set the first and last 4 columns of the offsets to 4095
! as is done in operations

  subroutine patchOffsets(ev_tab)

! ev_tab   : Handle to the input event table

    type(TableT), intent(in)                     :: ev_tab

    integer, parameter  :: patchWidth = 4
    type(DataSetT)      :: ev_set
    type(TableT)        :: off_tab
    type(ColumnT)       :: offx_col
    integer             :: windX0, windDX, windX1

! Must be done before the window keywords have been updated

    if (wasDone(block(ev_tab),changexy))  return
    call message(AppMsg,Verbose,"Patching the side offsets to 4095")

! Read the window size

    windX0 = int32Attribute(ev_tab,"WINDOWX0")
    windDX = int32Attribute(ev_tab,"WINDOWDX")
    windX1 = windX0 + windDX - 1

! Read the offsets

    ev_set = parent(ev_tab)
    if (.not. hasBlock(ev_set, offset_ext))  return

    off_tab = table(ev_set, offset_ext)
    offx_col = column(off_tab, "COLOFFST", Modify)
    offsetX => int16VectorData(offx_col)

! Set the borders to 4095

    offsetX(windX0:windX0+patchWidth-1,0) = 4095
    offsetX(windX1-patchWidth+1:windX1,0) = 4095

    call release(off_tab)

  end subroutine patchOffsets

! Get the offsets from CAL and write them to an extension

  subroutine getCalOffsets(ev_tab)

! ev_tab   : Handle to the input event table

    type(TableT), intent(in) :: ev_tab

    type(DataSetT)      :: ev_set
    type(TableT)        :: off_tab
    type(ColumnT)       :: offx_col, offy_col

    integer(kind=int16), dimension(:), pointer :: caloffX, caloffY
    integer             :: windX0, windDX, windX1, windY0, windDY, windY1

    call message(AppMsg,Sparse,"Getting the offsets from the CAL")

! Read the window size

    windX0 = int32Attribute(ev_tab,"WINDOWX0")
    windDX = int32Attribute(ev_tab,"WINDOWDX")
    windX1 = windX0 + windDX - 1
    windY0 = int32Attribute(ev_tab,"WINDOWY0")
    windDY = int32Attribute(ev_tab,"WINDOWDY")
    windY1 = windY0 + windDY - 1

! Read the offsets from the CAL

    call CAL_getMOSoffsets(caloffY,caloffX)
    if (size(caloffX) /= windDX .or. size(caloffY) /= windDY) then
       write(errstr,*) "The window size in ", trim(qualifiedName(ev_tab)),  &
                       " ", windDX, windDY, "is not the same as the size ",  &
                       "of the arrays returned by CAL_getMOSoffsets ",  &
                       size(caloffX), size(caloffY), ". Contact developer"
       call warning("getCalOffsets11")
       call CAL_releaseMemory(caloffX,caloffY)
       return
    endif

! Create the offsets extension

    ev_set = parent(ev_tab)
    off_tab = addTable(ev_set, offset_ext, 1, "Offsets from CAL")
    call setAttribute(off_tab, "DATATYPE", "OFFVAR.EL")
    offx_col = addColumn(off_tab, "COLOFFST", Integer16,  &
               comment="Row offset value", dimensions=(/x_size_window/))
    offy_col = addColumn(off_tab, "ROWOFFST", Integer16,  &
               comment="Column offset value", dimensions=(/y_size_window/))
    offsetX => int16VectorData(offx_col)
    offsetY => int16VectorData(offy_col)

! Set them to 100 by default

    offsetX = 100
    offsetY = 100

! Copy the offsets to the new extension
! The CAL provides them only within the window

    offsetX(windX0:windX1,0) = caloffX
    offsetY(windY0:windY1,0) = caloffY

    call CAL_releaseMemory(caloffX,caloffY)
    call release(off_tab)
    onoffvar = .true.

  end subroutine getCalOffsets

! Read the input frame file

  subroutine readFrames(frame_in, ev1_tab, ev_tab, fr_tab )

! frame_in  : Name of the input event file
! ev1_tab   : Handle to the input event table
! ev_tab    : Handle to the output event table
! fr_tab    : Handle to the input frame table

    character(len=*), intent(in) :: frame_in
    type(TableT), intent(in)     :: ev_tab, ev1_tab
    type(TableT), intent(out)    :: fr_tab
  
! Keywords to check for compatibility of frame and event files

    character(len=8), dimension(6), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","OBS_ID  ","EXP_ID  ","CCDID   ","CCDNODE "/)

    type(DataSetT)    :: fr_set
    integer           :: nframes
    logical           :: compatible, gatti_on

    call message(AppMsg,Verbose,"Reading the frame file ...")

    gatti_on = .false.
    fr_set = dataSet(frame_in, Read)
    if (hasBlock(fr_set,frame_ext)) then       
       fr_tab = table(fr_set, frame_ext)
       if (hasColumn(fr_tab,trim(fram)) .and. hasColumn(fr_tab,trim(time))  &
           .and. numberOfRows(fr_tab) > 1) then
          compatible = equalKeywords(attributable(fr_tab),  &
                                     attributable(ev1_tab), keycompat)
          if (compatible) then
             call message(AppMsg,Verbose,  &
                          "Frame file compatible with event file")
  
             call addFilename(block(ev_tab), frame_in, "Input frame file :")

! Propagate all keywords except DATATYPE

             call addAttributes(attributable(ev_tab),attributable(fr_tab))
             call setAttribute(ev_tab,attribute(ev1_tab,"DATATYPE"))
             call addAttributes(attributable(parent(ev_tab)),  &
                  attributable(parent(fr_tab)))
        
             if (hasAttribute(fr_tab,"GATTI_ON"))  &
                gatti_on = booleanAttribute(fr_tab,"GATTI_ON")

             if (.not. wasDone(block(fr_tab), frames_on)) then
                write(errstr,*) "the frame file did not go through emframes."
                call warning("readFrames14")
             endif

             fr_col = column(fr_tab, fram, Read)
             frameno => int32Data(fr_col)
             nframes = size(frameno)
             if (frameno(nframes-1)-frameno(0) /= nframes-1) then
                write(errstr,*) "There are missing or duplicate frames  ", &
                   "in the input frame table ", trim(qualifiedName(fr_tab)),  &
                   ". Contact developer of emframes."
                call error("readFrames01")
             endif
             time_col = column(fr_tab, time, Read)
             timetag => real64Data(time_col)

             frameflag = hasColumn(fr_tab,trim(flag))
             if (frameflag) then
                flag_col = column(fr_tab, flag, Read)
                quality => int16Data(flag_col)
             else if (oncutgti .or. outgatti) then
                oncutgti = .false.
                outgatti = .false.
                write(errstr,*) "Column ", flag,  &
                      " not present in ", trim(qualifiedName(fr_tab)),  &
                      ". CUT_GTI not applied and GATTI histograms not produced"
                call warning("readFrames10")
             endif

             call getNvalid(ev1_tab, frameno, nvalid, firstframe, lastframe,  &
                      frameoff, nonIncreasingFrame)   ! Always recompute NVALID

             onexpos = hasColumn(fr_tab,trim(crra))
             if (onexpos) then
                cr_col = column(fr_tab, crra, Read)
                crratio => real32Data(cr_col)
             else if (onputtime) then
                write(errstr,*) "Column ", crra, " not present in ",  &
                   trim(frame_in), ". EXPOSURE extension will have 1.0 ",  &
                   "everywhere. Run emframes with crdead_on=yes first"
                call warning("readFrames11")
             endif

             if (onspgatti) then
                 if (hasColumn(fr_tab,trim(gatv))) then
                    gatv_col = column(fr_tab, gatv, Read)
                    gattival => int16Data(gatv_col)
                    if (outgatti .and. .not. gatti_on) then
                       outgatti = .false.
                       call warning("readFrames18",  &
                      "GATTI off for this data. GATTI histograms not produced")
                    endif
                 else
                    onspgatti = .false.
                    write(errstr,*) "column ", gatv, " not present in ",  &
                          trim(frame_in), ". SPGATTI not applied. ",  &
                          "Run emframes with putgatti_on=yes first"
                    call warning("readFrames12")
                 endif
             endif
             return
          else
             write(errstr,*) "Frame file is not used"
             call warning("readFrames15")
          endif
       else if (numberOfRows(fr_tab) < 2) then
          if (numberOfRows(fr_tab) == 1) then
             call warning("readFrames19",  &
                        "The input frame file must contain at least 2 rows")
          else
             call warning("readFrames20", "The input frame file is empty")
          endif
       else
          write(errstr,*) "Columns ", fram, " and ", time,   &
                          " not present in ", trim(frame_in)
          call warning("readFrames16")
       endif
       call release(fr_tab)
    else
       write(errstr,*) "Extension ", frame_ext,  &
                       " not present in ", trim(frame_in)
       call warning("readFrames17")
    endif

    call warning("readFrames13",  &
                 "CUT_GTI, PUT_TIME and SP_GATTI will not be applied")
    frameread = .false.
    oncutgti  = .false.
    onspgatti = .false.
    call release(fr_set)

    call getNvalid(ev1_tab, frameno, nvalid, firstframe, lastframe,  &
                      frameoff, nonIncreasingFrame)

  end subroutine readFrames

! Read the input frame file of the second node (binodal)

  subroutine read2Frames(frame_in, ev1_tab, fr_tab)

! frame_in  : Name of the input event file
! ev1_tab   : Handle to the input event table
! fr_tab    : Handle to the output frame table

    character(len=*), intent(in) :: frame_in
    type(TableT), intent(in)     :: ev1_tab
    type(TableT), intent(out)    :: fr_tab

! Keywords to check for compatibility of frame and event files

    character(len=8), dimension(6), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","OBS_ID  ","EXP_ID  ","CCDID   ","CCDNODE "/)

    type(DataSetT)    :: fr_set
    type(ColumnT)     :: fr2_col, time2_col, flag2_col
    real(kind=double),   dimension(:), pointer :: timetag2
    integer(kind=int32), dimension(:), pointer :: frameno2
    integer(kind=int16), dimension(:), pointer :: nvalinter
    integer           :: nframes, firstframe2, lastframe2
    logical           :: compatible

    call message(AppMsg,Verbose,"Reading the secondary frame file ...")

    fr_set = dataSet(frame_in, Read)
    if (hasBlock(fr_set,frame_ext)) then       
       fr_tab = table(fr_set, frame_ext)
       if (hasColumn(fr_tab,trim(fram)) .and. hasColumn(fr_tab,trim(time))  &
           .and. numberOfRows(fr_tab) > 1) then
          compatible = equalKeywords(attributable(fr_tab),  &
                                     attributable(ev1_tab), keycompat)
          if (compatible) then
             call message(AppMsg,Verbose,  &
                          "Secondary frame file compatible with event file")

             if (.not. wasDone(block(fr_tab), frames_on)) then
                write(errstr,*) "The secondary frame file ",  &
                     trim(frame_in), " did not go through emframes."
                call warning("read2Frames14")
             endif

             fr2_col = column(fr_tab, fram, Read)
             frameno2 => int32Data(fr2_col)
             nframes = size(frameno2)
             if (frameno2(nframes-1)-frameno2(0) /= nframes-1) then
                write(errstr,*) "There are missing or duplicate frames in ", &
                   "the secondary frame table ", trim(qualifiedName(fr_tab)), &
                   ". BINODAL will not be applied."
                call warning("read2Frames11")
                onbinodal = .false.
                call release(fr_set)
             endif
             time2_col = column(fr_tab, time, Read)
             timetag2 => real64Data(time2_col)

             frameflag2 = hasColumn(fr_tab,trim(flag))
             if (frameflag2) then
                flag2_col = column(fr_tab, flag, Read)
                quality2 => int16Data(flag2_col)
             else 
                write(errstr,*) "Column ", flag,  &
                      " not present in ", trim(qualifiedName(fr_tab)),  &
                      ". Bad secondary frames will not be rejected for BINODAL"
                call warning("read2Frames10")
             endif

             call getNvalid(ev1_tab,frameno2,nvalid2,firstframe2,lastframe2, &
                     frameoff2,nonIncreasingFrame2)   ! Always recompute NVALID
             call binodal(timetag2)
             call release(fr2_col)
             call release(time2_col)

! In the case 'lastframe+offframe > lastframe2',  nvalid2 needs to be expanded
! with 0 values to the length (0:lastframe+offframe-frameoff2)

             if ( lastframe+offframe > lastframe2 ) then
                allocate(nvalinter(0:lastframe+offframe-frameoff2))
                nvalinter = 0
                nvalinter(0:lastframe2-frameoff2) = nvalid2
                deallocate(nvalid2)
                allocate(nvalid2(0:lastframe+offframe-frameoff2))
                nvalid2=nvalinter
                deallocate(nvalinter)
             endif

             return
          endif
       else if (numberOfRows(fr_tab) < 2) then
          if (numberOfRows(fr_tab) == 1) then
             write(errstr,*) "The secondary frame file ", trim(frame_in),  &
                             " must contain at least 2 rows"
             call warning("read2Frames19")
          else
             write(errstr,*) "The secondary frame file ", trim(frame_in),  &
                             " is empty"
             call warning("read2Frames20")
          endif
       else
          write(errstr,*) "Columns ", fram, " and ", time,   &
                          " not present in ", trim(frame_in)
          call warning("read2Frames16")
       endif
       call release(fr_tab)
    else
       write(errstr,*) "Extension ", frame_ext,  &
                       " not present in ", trim(frame_in)
       call warning("read2Frames17")
    endif

    call warning("read2Frames13", "BINODAL will not be applied")
    onbinodal = .false.
    call release(fr_set)

  end subroutine read2Frames

! Compute NVALID directly from events file

  subroutine getNvalid(ev1_tab, frameno, nvalid, firstframe, lastframe,  &
                      frameoff, nonIncreasingFrame)

! ev1_tab   : Handle to the input event table
! frameno   : Frame numbers in frame file
! nvalid    : Number of valid events per frame
! firstframe : Frame number corresponding to nvalid(0)
! lastframe : firstframe + size(nvalid) - 1
! frameoff  : firstframe - 1

    type(TableT), intent(in)  :: ev1_tab
    integer(kind=int32), dimension(:), pointer :: frameno, nvalid
    integer,      intent(out) :: firstframe, lastframe, frameoff
    logical,      intent(out) :: nonIncreasingFrame

! Flag values for frames

    integer(kind=int16) :: testflag
    integer(kind=int32),  dimension(:), pointer  :: framenumber
    type(ColumnT)  :: frame_col
    integer        :: nevents, nframes, iv, nf, lv
    integer        :: firsteventframe, lasteventframe 

    call message(AppMsg,Verbose,"Recomputing NVALID from the event list ...")

    nevents = numberOfRows(ev1_tab)
    frame_col = column(ev1_tab, fram, Read)
    framenumber => int32Data(frame_col)

!  Construct bit mask for testing the flags for the good frames
!  All bit values below flag_reject are OK. flag_hkgti also.

    testflag = flag_reject - 1 + flag_hkgti
    testflag = not(testflag)

! Define the first and last good frames from the frame file
! There may be 0 events in some frames

    if (frameread) then
       nframes = size(frameno)
       if (frameflag) then
          if (nevents > 0) then
             firsteventframe = framenumber(0)
             lasteventframe = framenumber(nevents-1)
          else
             firsteventframe = frameno(nframes-1)
             lasteventframe = frameno(1)
          endif
          do nf = 1,nframes-1
             if (firsteventframe <= frameno(nf) .or.  &
                 iand(quality(nf),testflag) == 0)  exit
          enddo
          firstframe = frameno(nf)
          do nf = nframes-1,1,-1
             if (lasteventframe >= frameno(nf) .or.  &
                 iand(quality(nf),testflag) == 0)  exit
          enddo
          lastframe = frameno(nf)
       else
          firstframe = frameno(1)
          lastframe = frameno(nframes-1)
       endif
       frameoff = frameno(1) - 1
    endif

    if (nevents > 0) then
       if (frameread) then
          if (framenumber(0) < firstframe) then
             write(errstr,*) "The first event in ",  &
                trim(qualifiedName(ev1_tab)), " has frame number ",  &
                framenumber(0), " smaller than the first frame ", firstframe, &
             " in the frame file. There must be a problem with the frame file."
             call error("getNvalid01")
          endif
          if (framenumber(nevents-1) > lastframe) then
             do iv = nevents-1,0,-1
                if (framenumber(iv) <= lastframe) exit
             enddo
             write(errstr,*) "The last", nevents-iv-1, " events in ",  &
                trim(qualifiedName(ev1_tab)), " have frame number up to",  &
                framenumber(nevents-1), " larger than the last frame ",  &
                lastframe, " in the frame file. They will not be propagated."
             call warning("getNvalid10")
             nevents = iv + 1
          endif
       else
          firstframe = framenumber(0)
          lastframe  = framenumber(nevents-1)
          frameoff = framenumber(0) - 1
       endif
    else if (.not. frameread) then
       firstframe = 0
       lastframe = -1
       frameoff = 0
    endif

    allocate(nvalid(0:lastframe-frameoff))
    nvalid = 0
    nf = 1
    lv = 0
    nonIncreasingFrame = .false.

! Recompute NVALID
! Check case of spurious frame number
! Avoid do while loops

    do iv = 0,nevents-1
       if (framenumber(iv) > lastframe) then
          write(errstr,*) "The frame number is too large at row", &
             iv+1, " in the input event file ", trim(qualifiedName(ev1_tab)), &
             ". Event will be rejected."
          call warning("getNvalid12")
          nonIncreasingFrame = .true.
       else if (framenumber(iv) > nf+frameoff) then   ! End of current frame
          nvalid(nf) = iv - lv
          nf = framenumber(iv) - frameoff
          lv = iv
       else if (framenumber(iv) < nf+frameoff) then
          write(errstr,*) "The frame number is non-increasing at row", &
             iv+1, " in the input event file ", trim(qualifiedName(ev1_tab)), &
             ". Event will be rejected."
          call warning("getNvalid11")
          nonIncreasingFrame = .true.
       endif
    enddo
    if (size(nvalid) > 0)  nvalid(nf) = nevents - lv    ! NVALID for last frame

    call release(frame_col)

  end subroutine getNvalid

! Weed out events with non-increasing frame number
! This is sometimes due to bad first frame. Should check for that.

  subroutine checkEventOrder(frame0, frame1, frameoff, evtbuf, nvalid,  &
                             orderbuf, nevents2)

! frame0 :   first frame in evtbuf
! frame1 :    last frame in evtbuf
! frameoff : offset between nvalid and frame number
! evtbuf :   event list structure
! nvalid :   number of events per frame
! orderbuf : list of good events
! nevents2 : number of good events

    integer,                           intent(in)    :: frame0,frame1,frameoff
    type(sas_evt),       dimension(:), intent(in)    :: evtbuf
    integer(kind=int32), dimension(:), pointer       :: nvalid
    integer(kind=int32), dimension(:), intent(inout) :: orderbuf
    integer,                           intent(out)   :: nevents2

    integer :: nnoninc, nevents3, count0, count1, irow, jrow, order

    nnoninc = 0
    nevents3 = 0
    count0 = 1
    do irow = frame0-frameoff,frame1-frameoff  ! Loop over frames
       if (nvalid(irow) == 0) cycle            ! No event in that frame
       count1 = count0 + nvalid(irow) - 1
       do jrow = count0, count1
          order = orderbuf(jrow)
          if (evtbuf(order)%frame-frameoff == irow) then
             nevents3 = nevents3 + 1
             orderbuf(nevents3) = order
          else
             nvalid(irow) = nvalid(irow) - 1
             nnoninc = nnoninc + 1
          endif
       enddo
       count0 = count1 + 1
    enddo
    nevents2 = nevents3
    if (nnoninc > 0) then
       write(errstr,*) nnoninc,  &
            " events rejected for wrong frame number"
       call message(AppMsg,Sparse)
    endif

  end subroutine checkEventOrder

! Read the required columns into global variables

  subroutine readColumns(ev_tab)

! ev_tab    : Handle to the input/output event table

    type(TableT), intent(in)   :: ev_tab

    call message(AppMsg,Verbose,"Reading the event file ...")

    if (oncutgti .or. onputtime .or. onspgatti) then
       frev_col = column(ev_tab, fram, Read)
       frameev => int32Data(frev_col)
    endif

    if (onputxy .or. onspgatti) then
       rawx_col = column(ev_tab, rawx, Modify)
       posx => int16Data(rawx_col)
    endif

    if (rowreject .or. (onputtime .and. (timeMode.or.ctimeMode)) .or.  &
        ((onputxy.or.onspgatti) .and. (imageMode.or.rimagMode))) then
       rawy_col = column(ev_tab, rawy, Modify)
       posy => int16Data(rawy_col)
    endif

    if (oncutgti .or. onputxy .or. onspgatti .or. onputtime .or.  &
        rowreject .or. e3reject) then
       if (hasColumn(ev_tab, EVATT_COLUMN_NAME)) then
          flev_col = column(ev_tab, EVATT_COLUMN_NAME, Modify)
       else
          flev_col = addColumn(ev_tab, EVATT_COLUMN_NAME, Integer32, "",  &
                       comment="quality flag")
       endif
       qualev => int32Data(flev_col)
    endif

  end subroutine readColumns

! Release the columns

  subroutine release_columns

    call message(AppMsg,Verbose,"Releasing the columns ...")

! Event file

    if (onputxy .or. onspgatti)  call release(rawx_col)
    if (rowreject .or.(onputtime .and. (timeMode.or.ctimeMode)) .or.  &
        ((onputxy.or.onspgatti) .and. (imageMode.or.rimagMode)))  &
        call release(rawy_col)
    if (oncutgti .or. onputtime .or. onspgatti) call release(frev_col)
    if (oncutgti .or. onputxy .or. onspgatti .or. onputtime .or.  &
        rowreject .or. e3reject)  call release(flev_col)

! Frame file

    if (frameread) then
       call release(fr_col)
       call release(time_col)
       if (frameflag)  call release(flag_col)
       if (onexpos)    call release(cr_col)
       if (onspgatti)  call release(gatv_col)
    endif
    deallocate(nvalid)

  end subroutine release_columns

! Conversion from date format to XMM time in seconds
! Returns XMM time for 01/01/2000 when error

  real(double) function getXmmDate(ev_tab)

! ev_tab: Handle to the input table

    type(TableT), intent(in) :: ev_tab

    character(len=80) :: dateobs
    real(kind=double) :: mjd, mjd0
    real              :: dj, rsec
    integer           ::  year, month, day, hour, min, sec, status

! Set date arbitrarily to 01/00/2000

    getXmmDate = 2.d0 * 365.25d0 * 24.d0 * 3600.d0
    if (.not. hasAttribute(ev_tab,"DATE-OBS")) then
       write(errstr,*) "No DATE-OBS keyword in ", trim(qualifiedName(ev_tab))
       call warning("getXmmDate12")
       return
    endif
    dateobs = stringAttribute(ev_tab,"DATE-OBS")
    if (len_trim(dateobs) < 19) then
       write(errstr,*) "Bad format DATE-OBS keyword ", trim(dateobs),  &
                       " in ", trim(qualifiedName(ev_tab)),  &
                       ". Should be yyyy-mm-ddThh:mm:ss"
       call warning("getXmmDate13")
       return
    endif
    if (dateobs( 5: 5) /= "-" .or. dateobs( 8: 8) /= "-" .or.  &
        dateobs(11:11) /= "T" .or.  &
        dateobs(14:14) /= ":" .or. dateobs(17:17) /= ":") then
       write(errstr,*) "Bad format DATE-OBS keyword ", trim(dateobs),  &
                       " in ", trim(qualifiedName(ev_tab)),  &
                       ". Should be yyyy-mm-ddThh:mm:ss"
       call warning("getXmmDate13")
       return
    endif
    read(dateobs,100) year, month, day, hour, min, sec
100 format(I4,1x,I2,1x,I2,1x,I2,1x,I2,1x,I2)
    call sla_cldj(1998,01,01,mjd0,status)
    call sla_cldj(year,month,day,mjd,status)
    if (status > 0) then
       write(errstr,*) "Error", status, " in sla_cldj for DATE-OBS = ",  &
          trim(dateobs), " in ", trim(qualifiedName(ev_tab)),  &
          ". Should be yyyy-mm-ddThh:mm:ss"
       call warning("getXmmDate11")
       return
    endif
    rsec = sec
    call sla_ctf2d(hour,min,rsec,dj,status)
    if (status > 0) then
       write(errstr,*) "Error", status, " in sla_ctf2d for DATE-OBS = ",  &
          trim(dateobs), " in ", trim(qualifiedName(ev_tab)),  &
          ". Should be yyyy-mm-ddThh:mm:ss"
       call warning("getXmmDate14")
       return
    endif

! Convert from days to seconds

    getXmmdate = (mjd - mjd0 + dj) * 24.d0 * 3600.d0

  end function getXmmDate


! This routine computes offframe which is the difference in framenumber between
! frames of the 2 different nodes which have the same time

  subroutine binodal(timetag2)

! timetag2 : time of frames for secondary node

    real(kind=double),   dimension(:), pointer :: timetag2

    real(kind=double) :: timefirst, timelast
    integer :: fr
    logical :: ok

! Time corresponding to firstframe
    
    timefirst = timetag( firstframe - frameoff )

! Look for the frame in the other node the time of which is closer

    ok = .false.
    do fr = 0, size(timetag2)-1 

! Absolute time comparison because we know the 2 nodes are driven by the
! same sequencer in the EMAE

       if ( (timefirst-timetag2(fr)) < .01 ) then
          ok = .true.
          exit
       endif
    enddo

    if (ok) then

       offframe = frameoff2+fr - firstframe

! Check that this difference remains constant for the whole exposure

       timelast= timetag( lastframe - frameoff )
       ok = .false.
       do fr = 0, size(timetag2)-1 

! Absolute time comparison because we know the 2 nodes are driven by the
! same sequencer in the EMAE

          if ( (timelast-timetag2(fr)) < .01 ) then
             ok = .true.
             exit
          endif
       enddo
       if (ok) then
        if ( offframe /= (frameoff2 + fr - lastframe) ) then
           write(errstr,*) &
                "Frame difference not constant within the whole exposure"
           call error("binodal02")

        endif
       endif

    else
       write(errstr,*) &
            "Can't find a frame in the 2nd node with a time close to ", &
            timefirst, &
            " which is the time of the first valid frame in 1st frame file"
       call error("binodal01")     

    endif

  end subroutine binodal

! Initialize the CAL and the image size

  subroutine initCAL(ev_tab)

! ev1_tab : handle to the input table

    type(TableT), intent(in) :: ev_tab

    type(ModeParameterDataType)  :: modeParameters

    call CAL_setState(block(ev_tab))

! Get under/overscans

    call CAL_getModeParameters(modeParameters)
    underx = modeParameters%underscanX
    overx = modeParameters%overscanX
    undery = modeParameters%underscanY
    overy = modeParameters%overscanY
    x_size_window = EMOS_MAX_X + underx + overx
    if (.not. timeMode) then 
       y_size_window = EMOS_MAX_Y + undery + overy
    else
       y_size_window = 1024 + 1  ! See cutBad() (shift +2 in Y)
    endif
    x_size_window_bin = EMOS_MAX_X/2 + underx + overx

  end subroutine initCAL

! Parameter reading

  subroutine getParamValues(ev1_tab, neventsin, off_tab, event_out, frame_in, &
                      event2, frame2, gattifile)

! ev1_tab   : Handle to the input event table
! neventsin : Number of events in input
! off_tab   : Handle to the offset table
! event_out : Name of the output event file
! frame_in  : Name of the input frame file
! event2    : Name of the second input event file (binodal)
! frame2    : Name of the second input frame file (binodal)
! gattifile : Name of the output GATTI histogram file

    character(len=*), intent(out) :: event_out, frame_in,  &
                                     event2, frame2, gattifile
    type(TableT), intent(out)     :: off_tab, ev1_tab
    integer, intent(out)          :: neventsin

    character(len=1024)                            :: event_in, instname
    character(len=1024), allocatable, dimension(:) :: offvar
    integer                                        :: numoffvar, nellipse
    integer                                        :: inst, ccdnum

    call message(AppMsg,Verbose,"Reading the parameters ...")

! Parameters related to the event files

    event_in = stringParameter(inevent)
    write(errstr,*) "Input event file : ",trim(event_in)
    call message(AppMsg,Verbose)

    overevent = .not. booleanParameter(newoutput)
    if (overevent) then
       write(errstr,*) "The input event file will be overwritten"
       call message(AppMsg,Verbose)
    else
       event_out = stringParameter(outevent)
       if (trim(event_out) == trim(event_in)) then
          write(errstr,*) "The input and output event files cannot have ",  &
                          "the same name. Use ",newoutput,"=no to overwrite."
          call error("getParamValues03")
       endif
       write(errstr,*) "Output event file : ",trim(event_out)
       call message(AppMsg,Verbose)
    endif

! Parameters related to the offset/variance file

    onoffvar = booleanParameter(withoffvar)
    numoffvar = sizeListParam(offvarfile)
    if (onoffvar .and. numoffvar < 1) then
       write(errstr,*) "Empty offset/variance list, though ",withoffvar,"=yes"
       call warning("getParamValues16")
       onoffvar = .false.
    endif
    if (onoffvar)  then
       allocate( offvar(numoffvar) )
       call readListParam(offvarfile,offvar,numoffvar)
    else
       numoffvar = 0                 ! It is neccessary to allocate offvar
       allocate( offvar(numoffvar) ) ! since it is an argument of the next call
    endif

    call openEventsIn(event_in,offvar,off_tab,ev1_tab,neventsin)

    onevrec    = booleanParameter(evrec_on)
    oncutbad   = booleanParameter(cutbad_on)
    ondiago    = booleanParameter(diago_on)
    onbinodal  = booleanParameter(twonodes)
    if (onevrec .and. .not. fromodf) then
       onevrec = .false.
       write(errstr,*) "The ",evrec_on,"=yes option is allowed only ",  &
          "on an event file straight from the ODF. It will not be applied."
       call warning("getParamValues17")
    endif
    if (oncutbad .and. .not. hasBlock(parent(ev1_tab), badpix_ext)) then
       oncutbad = .false.
       write(errstr,*) "The input event file has no ",badpix_ext,  &
                       " extension. CUT_BAD will not be applied."
       call warning("getParamValues10")
    endif
    if (oncutbad .and. .not. fromodf) then
       oncutbad = .false.
       write(errstr,*) "The ",cutbad_on,"=yes option is allowed only ",  &
          "on an event file straight from the ODF. It will not be applied."
       call warning("getParamValues12")
    endif
    if (ondiago .and. .not. fromodf) then
       ondiago = .false.
       write(errstr,*) "The ",diago_on,"=yes option is allowed only ",  &
          "on an event file straight from the ODF. It will not be applied."
       call warning("getParamValues13")
    endif
    if (onbinodal .and. .not. fromodf) then
       onbinodal = .false.
       write(errstr,*) "The ",twonodes,"=yes option is allowed only ",  &
          "on an event file straight from the ODF. It will not be applied."
       call warning("getParamValues14")
    endif

! EV_REC does not make sense in IMAGING mode unless BINODAL or CUT_BAD

    if (onevrec.and.imageMode .and. .not.(oncutbad.or.onbinodal)) then
       onevrec = .false.
       write(errstr,*) "The ",evrec_on,"=yes option is useless in IMAGING ",  &
          "mode unless ",cutbad_on,"=yes or ",twonodes,"=yes. Not done."
       call warning("getParamValues21")
    endif

! EV_REC and CUT_BAD do not make sense in Compressed Timing mode

    if (ctimeMode .and. (onevrec.or.oncutbad)) then
       onevrec = .false.
       oncutbad = .false.
       write(errstr,*) "The ", evrec_on, "=yes and ",cutbad_on, "=yes ",  &
          "options cannot be applied in COMPRESSED TIMING mode. Not done."
       call warning("getParamValues19")
    endif

! BINODAL does not make sense unless EV_REC

    if (onbinodal .and. .not.onevrec) then
       onbinodal = .false.
       write(errstr,*) "The ",twonodes,"=yes option requires ",evrec_on,  &
                       "=yes. Not done."
       call warning("getParamValues22")
    endif

! DIAGO does not make sense in TIMING mode

    if (ondiago .and. (timeMode.or.ctimeMode)) then
       ondiago = .false.
       write(errstr,*) "The ",diago_on,"=yes option may be applied only ",  &
                       "in IMAGING mode. Not done."
       call warning("getParamValues11")
    endif
    if (ondiago) then
       call message(AppMsg,Verbose,"DIAGO will be applied")
       onsifluor = booleanParameter(keepsifluor)
       if (onsifluor)  call message(AppMsg,Verbose,  &
          "Diagonal events associated with Si fluorescence will not be split.")
    endif

! Parameters related to the frame file

    frameread = booleanParameter(withframe)
    if (frameread) then
       frame_in = stringParameter(framefile)
       write(errstr,*) "Input frame file : ",trim(frame_in)
       call message(AppMsg,Verbose)
       oncutgti   = booleanParameter(cutgti_on)
       onspgatti  = booleanParameter(spgatti_on)
    else
       oncutgti   = .false.
       onspgatti  = .false.
    endif

    onputtime  = booleanParameter(puttime_on)
    timerandom = .false.
    if (onputtime)  timerandom = booleanParameter(randomtime)

! Time randomisation makes sense only in IMAGING mode

    if ((timeMode .or. ctimeMode) .and. timerandom) then
       timerandom = .false.
       write(errstr,*) "The ",randomtime,"=yes option may be applied only ",  &
                       "in IMAGING mode. Not done."
       call warning("getParamValues18")
    endif

    if (onbinodal) then
       event2 = stringParameter(inevent2)
       write(errstr,*) "BINODAL will be applied. ",  &
                       "Second input event file: ",trim(event2)
       call message(AppMsg,Verbose)
       if (frameread) then
          frame2 = stringParameter(framefile2)
          write(errstr,*) "Second input frame file: ",trim(frame2)
          call message(AppMsg,Verbose)
       endif
    endif

    onputxy    = booleanParameter(changexy)
    xyrandom = .false.
    if (onputxy)  xyrandom = booleanParameter(randomxy)

    nmaxevents = intParameter(blocksize)

! Row rejection makes sense only in IMAGING mode

    rowreject  = booleanParameter(badrows_on)
    if ( rowreject) then
       maxevrow = intParameter(threshrow)
       write(errstr,*) "Rows with more than", maxevrow,  &
                       " events in a single frame will be rejected"
       call message(AppMsg,Verbose)
    endif

! Rejection on E3 makes sense only in strict IMAGING mode

    e3reject = booleanParameter(bade3_on)
    if ( e3reject ) then
       if (.not. imageMode) then
          e3reject = .false.
          write(errstr,*) "The ",bade3_on,"=yes option may be applied ",  &
                          "only in IMAGING mode. Not done."
          call warning("getParamValues20")
       else
          nellipse = parameterCount(e3threshold)
          if (nellipse .eq. 14) then
             instname = changeCase(stringAttribute(ev1_tab,"INSTRUME"),2)
             if (trim(instname) .eq. "EMOS1") then
                inst = 1
             elseif (trim(instname) .eq. "EMOS2") then
                inst = 2
             else
                write(errstr,*) "INSTRUME = ", trim(instname),  &
                                " was not recognized. Should be EMOS1 or EMOS2"
                call error("getParamValues05")
             endif
             ccdnum = int32Attribute(ev1_tab,"CCDID") - 1 + 7*(inst-1)
             e3min = intParameter(e3threshold,ccdnum)
          else if (nellipse .eq. 1) then
             e3min = intParameter(e3threshold,0)
          else
             write(errstr,*) "The number of thresholds in ",  &
                             e3threshold, nellipse, " should be 1 or 14"
             call error("getParamValues04")
          endif
          write(errstr,*) "Events with E3 larger than", e3min,  &
                          " will be rejected"
          call message(AppMsg,Verbose)
       endif
    endif

! SP_GATTI may not be applied in Compressed TIMING mode (no PATTERN)

    if (onspgatti) then
       if (ctimeMode) then
          onspgatti = .false.
          write(errstr,*) "The ",spgatti_on,"=yes option may not be ",  &
                          "applied in Compressed TIMING mode. Not done."
          call warning("getParamValues15")
       else
          call message(AppMsg,Verbose,"SP_GATTI will be applied")
          outgatti = booleanParameter(writegatti)
          if (outgatti) then
             gattifile = stringParameter(gattispec)
             write(errstr,*) "Output GATTI histograms file : ",trim(gattifile)
             call message(AppMsg,Verbose)
          endif
       endif
    endif

    if (onevrec)   call message(AppMsg,Verbose,"EV_REC will be applied")
    if (oncutbad)  call message(AppMsg,Verbose,"CUT_BAD will be applied")
    if (onputxy)   call message(AppMsg,Verbose,"PUT_XY will be applied")
    if (xyrandom)  call message(AppMsg,Verbose,"DETX/DETY will be randomized")
    if (onputtime) call message(AppMsg,Verbose,"PUT_TIME will be applied")
    if (timerandom) call message(AppMsg,Verbose,"TIME will be randomized")
    if (oncutgti)  call message(AppMsg,Verbose,"CUT_GTI will be applied")

! store node number in global variable

    ccdnode = int32Attribute(ev1_tab,"CCDNODE")

  end subroutine getParamValues

end module emevents_mod
